import logging

from click import help_option, argument, option, Choice
from rich.console import Console
from rich.table import Table

from duckcp.boot import app
from duckcp.repository import RepositoryKind
from duckcp.service import transformer_service

logger = logging.getLogger(__name__)


@app.group(help='管理迁移')
@help_option('-h', '--help', help='展示帮助信息')
def transformer():
    pass


@transformer.command('create', help='创建迁移')
@argument('name', metavar='NAME')
@option('-s', '--source', metavar='REPOSITORY', required=True, help='来源仓库')
@option('-t', '--target', metavar='REPOSITORY', required=True, help='目标仓库')
@option('-o', '--storage', metavar='STORAGE', required=True, help='目标存储单元')
@option('-f', '--script', metavar='FILE', required=True, help='迁移脚本')
@help_option('-h', '--help', help='展示帮助信息')
def transformer_create(name: str, source: str, target: str, storage: str, script: str):
    logger.debug('name=%s, source=%s, target=%s, storage=%s, script=%s', name, source, target, storage, script)
    transformer_service.transformer_create(name, source, target, storage, script)


@transformer.command('update', help='更新迁移信息')
@argument('name', metavar='NAME')
@option('-s', '--source', metavar='REPOSITORY', help='来源仓库')
@option('-t', '--target', metavar='REPOSITORY', help='目标仓库')
@option('-o', '--storage', metavar='STORAGE', help='目标存储单元')
@option('-f', '--script', metavar='FILE', help='迁移脚本')
@help_option('-h', '--help', help='展示帮助信息')
def transformer_update(name: str, source: str, target: str, storage: str, script: str):
    logger.debug('name=%s, source=%s, target=%s, storage=%s, script=%s', name, source, target, storage, script)
    transformer_service.transformer_update(name, source, target, storage, script)


@transformer.command('delete', help='删除迁移；更新作业')
@argument('name', metavar='NAME')
@help_option('-h', '--help', help='展示帮助信息')
def transformer_delete(name: str):
    logger.debug('name=%s', name)
    transformer_service.transformer_delete(name)


@transformer.command('list', help='列出所有迁移')
@option('--source-kind', type=Choice(RepositoryKind.codes()), help='来源仓库类型')
@option('--source-repository', metavar='REPOSITORY', help='来源仓库名称')
@option('--target-kind', type=Choice(RepositoryKind.codes()), help='目标仓库类型')
@option('--target-repository', metavar='REPOSITORY', help='目标仓库名称')
@option('--target-storage', metavar='STORAGE', help='目标存储单元')
@help_option('-h', '--help', help='展示帮助信息')
def transformer_list(
        source_kind: str,
        source_repository: str,
        target_kind: str,
        target_repository: str,
        target_storage: str,
):
    table = Table(title='迁移列表')
    table.add_column('名称', no_wrap=True)
    table.add_column('来源类型')
    table.add_column('来源仓库')
    table.add_column('目标类型')
    table.add_column('目标仓库')
    table.add_column('存储单元')
    table.add_column('迁移脚本')
    table.add_column('关联任务', justify='right')
    for row in transformer_service.transformer_list(source_kind, source_repository, target_kind, target_repository, target_storage):
        table.add_row(
            row.code,
            row.source_repository_kind,
            row.source_repository_code,
            row.target_repository_kind,
            row.target_repository_code,
            row.target_storage_code,
            row.script_file,
            str(row.tasks),
        )

    console = Console()
    console.print(table)


@transformer.command('execute', help='执行迁移')
@argument('name', metavar='NAME')
@help_option('-h', '--help', help='展示帮助信息')
def transformer_execute(name: str):
    logger.debug('name=%s', name)
    transformer_service.transformer_execute(name)



"""
HTTP客户端帮助函数。
"""
import logging
from typing import Any
from urllib.parse import urlparse, parse_qs, urlencode, urlunparse
from urllib.request import urlopen, Request

from duckcp.helper.serialization import json_decode, json_encode

logger = logging.getLogger(__name__)


class HttpClientError(Exception):
    """
    HTTP客户端执行失败时，抛出本异常。
    """

    def __init__(self, url: str, message: str):
        super().__init__(f'请求{url}失败：{message}')


def request(body: Request, timeout: float = 10) -> Any:
    with urlopen(body, timeout=timeout) as response:
        if response.status // 100 == 2:  # 2xx
            data = response.read().decode()
            return json_decode(data)
        else:
            raise HttpClientError(body.full_url, response.status)


def _append_search(url: str, query: dict) -> str:
    """
    将查询条件合并到URL中。
    """
    if query is not None:
        uri = urlparse(url)
        query = {**parse_qs(uri.query), **query}  # 合并参数，并且用新参数覆盖已有参数
        search = urlencode(query, doseq=True)
        return str(urlunparse(uri._replace(query=search)))
    else:
        return url


def get(url: str, headers: dict = None, query: dict = None, timeout: float = 10) -> Any:
    """
    发起HTTP GET请求。
    """
    logger.debug('url=%s, headers=%s, params=%s, timeout=%s', url, headers, query, timeout)

    if headers is None:
        headers = {}
    url = _append_search(url, query)

    return request(Request(url, method='GET', headers=headers), timeout)


def post(url: str, headers: dict = None, query: dict = None, params: dict = None, timeout: float = 10) -> Any:
    """
    发起HTTP POST请求。
    """
    logger.debug('url=%s, headers=%s, params=%s, timeout=%s', url, headers, params, timeout)
    if headers is None:
        headers = {}
    if 'Content-Type' not in headers:
        headers['Content-Type'] = 'application/json; charset=utf-8'
    url = _append_search(url, query)
    data = json_encode(params).encode() if params is not None else None

    return request(Request(url, method='POST', headers=headers, data=data), timeout)



import logging

import click

from duckcp.boot import app
from duckcp.service import meta_service

logger = logging.getLogger(__name__)


@app.group(help='管理元信息数据库')
@click.help_option('-h', '--help', help='展示帮助信息')
def meta():
    pass


@meta.command('delete', help='删除元信息数据库')
@click.help_option('-h', '--help', help='展示帮助信息')
def meta_delete():
    meta_service.meta_delete()


@meta.command('create', help='创建元信息数据库')
@click.option('-f', '--force', is_flag=True, help='强制重建')
@click.help_option('-h', '--help', help='展示帮助信息')
def meta_create(force: bool):
    if force:
        meta_service.meta_delete()
    meta_service.meta_create()



"""
飞书开放接口
"""
import logging
from datetime import datetime, timedelta
from typing import TypedDict

from duckcp.helper import http
from duckcp.typing.authentication_token_type import AuthenticationToken

OPEN_API = 'https://open.feishu.cn/open-apis'
ACCESS_TOKEN_API = f'{OPEN_API}/auth/v3/tenant_access_token/internal/'

logger = logging.getLogger(__name__)


class FeiShuError(Exception):
    """
    飞书多维表格接口执行失败时，抛出本异常。
    """

    def __init__(self, api: str, message: str):
        super().__init__(f'{api}请求失败：{message}')


class CredentialResponse(TypedDict):
    """
    飞书应用凭证响应。
    """
    code: int  # 错误编码。
    msg: str  # 错误消息。
    expire: int  # 有效时长；单位秒。
    tenant_access_token: str  # 租户级别凭证。


def tenant_access_token(token: AuthenticationToken) -> tuple[str, datetime]:
    """
    获得租户级别的访问凭证
    @param token: 认证信息
    """
    logger.debug('token=%s', token)
    response: CredentialResponse = http.post(ACCESS_TOKEN_API, params={
        'app_id': token['access_key'],
        'app_secret': token['access_secret'],
    })
    if response['code'] == 0:
        access_token = response['tenant_access_token']
        expired_at = datetime.now() + timedelta(seconds=response['expire'] - 30)
        return access_token, expired_at
    else:
        raise FeiShuError('刷新凭证', response.get('msg', ''))



import logging
from typing import Any

from click import help_option, option, argument, INT, Choice
from rich.console import Console
from rich.table import Table

from duckcp.boot import app
from duckcp.helper.click import JSON
from duckcp.repository import RepositoryKind
from duckcp.service import storage_service

logger = logging.getLogger(__name__)


@app.group(help='管理存储单元')
@help_option('-h', '--help', help='展示帮助信息')
def storage():
    pass


@storage.command('create', help='创建存储单元')
@argument('name', metavar='NAME')
@option('-r', '--repository', metavar='REPOSITORY', help='所属仓库')
# Postgres；DuckDB；ODPS
@option('--catalog', metavar='CATALOG', help='目录；用于[postgres；duckdb；odps]')
# Postgres；DuckDB；ODPS；SQLite
@option('--schema', metavar='SCHEMA', help='模式；用于[postgres；duckdb；odps；sqlite]')
# Postgres；DuckDB；ODPS；SQLite；BiTable
@option('--table', metavar='TABLE', help='表；用于[postgres；duckdb；odps；sqlite；bitable]')
# BiTable
@option('--document', metavar='DOCUMENT', help='多维表格文档；用于[bitable]')
# File
@option('--file', metavar='FILE', help='文件名；用于[file]')
@option('--format', metavar='FORMAT', type=Choice(['csv', 'parquet', 'json']), help='文件格式；用于[file]')
@option('--compression', metavar='ALGORITHM', type=Choice(['gzip', 'zstd', 'snappy', 'brotli', 'lz4', 'lz4_raw']), help='压缩算法；用于[file]')
@option('--compression-level', metavar='LEVEL', type=INT, help='压缩等级；用于[file][parquet]')
@option('--parquet-version', metavar='VERSION', type=Choice(['V1', 'V2']), help='Parquet文件版本；用于[file][parquet]')
@option('--field-ids', metavar='JSON', type=JSON, help='字段的编号；用于[file][parquet]')
@option('--row-group-size', metavar='SIZE', type=INT, help='最大行数；用于[file][parquet]')
@option('--row-group-size-bytes', metavar='SIZE', type=INT, help='每行组最大字节数；用于[file][parquet]')
@option('--row-group-per-file', metavar='SIZE', type=INT, help='每个文件最大的行组数；用于[file][parquet]')
@option('--header/--no-header', is_flag=True, default=None, help='是否包含字段名；用于[file][csv]')
@option('--delimiter', metavar='DELIMITER', help='字段分隔符；用于[file][csv]')
@option('--quote-char', metavar='CHAR', help='引号字符；用于[file][csv]')
@option('--escape-char', metavar='CHAR', help='转义字符；用于[file][csv]')
@option('--null-literal', metavar='LITERAL', help='空值字面量；用于[file][csv]')
@option('--force-quote', metavar='COLUMN', multiple=True, help='给指定列强制添加引号；用于[file][csv]')
@option('--prefix', metavar='PREFIX', help='前置内容；用于[file][csv]')
@option('--suffix', metavar='SUFFIX', help='后置内容；用于[file][csv]')
@option('--date-format', metavar='FORMAT', help='日期格式；用于[file][csv；json]')
@option('--timestamp-format', metavar='FORMAT', help='时间戳格式；用于[file][csv；json]')
@option('--array/--no-array', is_flag=True, default=None, help='写入JSON数组；用于[file][json]')
@option('--per-thread-output', is_flag=True, default=None, help='是否每个线程写入独立文件；用于[file][thread]')
@option('--file-size-bytes', metavar='BYTES', type=INT, help='每个独立文件最大字节数；用于[file][thread]')
@option('--partition-by', metavar='COLUMN', multiple=True, help='分区列；用于[file][partition]')
@option('--filename-pattern', metavar='PATTERN', help='文件名模式；用于[file][partition]')
@option('--file-extension', metavar='EXTENSION', help='扩展名；用于[file][partition]')
@option('--write-partition-columns', is_flag=True, default=None, help='是否包含分区列；用于[file][partition]')
@option('--use-tmp-file', is_flag=True, default=None, help='若原始文件存在，是否先写入临时文件；用于[file][partition]')
@option('--delete-before-write', is_flag=True, default=None, help='写入之前删除整个目录；用于[file][partition]')
@option('--overwrite/--no-overwrite', is_flag=True, default=None, help='覆盖已存在的文件；用于[file][partition]')
@option('--append/--no-append', is_flag=True, default=None, help='追加到已存在的文件；用于[file][partition]')
@option('--preserve-order', is_flag=True, default=None, help='是否保留原始顺序；用于[file]')
# Others
@help_option('-h', '--help', help='展示帮助信息')
def storage_create(
        name: str,
        repository: str,
        # Postgres；DuckDB；ODPS
        catalog: str,  # 目录；用于[postgres；duckdb；odps]
        # Postgres；DuckDB；ODPS；SQLite
        schema: str,  # 模式；用于[postgres；duckdb；odps；sqlite]
        # Postgres；DuckDB；ODPS；SQLite；BiTable
        table: str,  # 表；用于[postgres；duckdb；odps；sqlite；bitable]
        # BiTable
        document: str,  # 多维表格文档；用于[bitable]
        # File
        file: str,  # 文件名；用于[file]
        format: str,  # 文件格式；用于[file]
        compression: str,  # 压缩算法；用于[file]
        compression_level: int,  # 压缩等级；用于[file][parquet]
        parquet_version: str,  # Parquet文件版本；用于[file][parquet]
        field_ids: dict[str, Any],  # 字段的编号；用于[file][parquet]
        row_group_size: int,  # 最大行数；用于[file][parquet]
        row_group_size_bytes: int,  # 每行组最大字节数；用于[file][parquet]
        row_group_per_file: int,  # 每个文件最大的行组数；用于[file][parquet]
        header: bool,  # 是否包含字段名；用于[file][csv]
        delimiter: str,  # 字段分隔符；用于[file][csv]
        quote_char: str,  # 引号字符；用于[file][csv]
        escape_char: str,  # 转义字符；用于[file][csv]
        null_literal: str,  # 空值字面量；用于[file][csv]
        force_quote: list[str],  # 给指定列强制添加引号；用于[file][csv]
        prefix: str,  # 前置内容；用于[file][csv]
        suffix: str,  # 后置内容；用于[file][csv]
        date_format: str,  # 日期格式；用于[file][csv；json]
        timestamp_format: str,  # 时间戳格式；用于[file][csv；json]
        array: bool,  # 写入JSON数组；用于[file][json]
        per_thread_output: bool,  # 是否每个线程写入独立文件；用于[file][thread]
        file_size_bytes: int,  # 每个独立文件最大字节数；用于[file][thread]
        partition_by: list[str],  # 分区列；用于[file][partition]
        filename_pattern: str,  # 文件名模式；用于[file][partition]
        file_extension: str,  # 扩展名；用于[file][partition]
        write_partition_columns: bool,  # 是否包含分区列；用于[file][partition]
        use_tmp_file: bool,  # 若原始文件存在，是否先写入临时文件；用于[file][partition]
        delete_before_write: bool,  # 写入之前删除整个目录；用于[file][partition]
        overwrite: bool,  # 覆盖已存在的文件；用于[file][partition]
        append: bool,  # 追加到已存在的文件；用于[file][partition]
        preserve_order: bool,  # 是否保留原始顺序；用于[file]
):
    logger.debug(
        'name=%s, repository=%s, catalog=%s, schema=%s, table=%s, document=%s, file=%s, format=%s, compression=%s, compression_level=%s, parquet_version=%s, field_ids=%s, row_group_size=%s, row_group_size_bytes=%s, row_group_per_file=%s, header=%s, delimiter=%s, quote_char=%s, escape_char=%s, null_literal=%s, force_quote=%s, prefix=%s, suffix=%s, date_format=%s, timestamp_format=%s, array=%s, per_thread_output=%s, file_size_bytes=%s, partition_by=%s, filename_pattern=%s, file_extension=%s, write_partition_columns=%s, use_tmp_file=%s, delete_before_write=%s, overwrite=%s, append=%s, preserve_order=%s',
        name, repository, catalog, schema, table, document,
        file, format, compression, compression_level,
        parquet_version, field_ids, row_group_size, row_group_size_bytes, row_group_per_file,
        header, delimiter, quote_char, escape_char, null_literal, force_quote, prefix, suffix,
        date_format, timestamp_format, array,
        per_thread_output, file_size_bytes,
        partition_by, filename_pattern, file_extension, write_partition_columns, use_tmp_file,
        delete_before_write, overwrite, append, preserve_order,
    )
    storage_service.storage_create(repository, name, {
        'catalog': catalog,
        'schema': schema,
        'table': table,
        'document': document,
        'file': file,
        'format': format,
        'compression': compression,
        'compression_level': compression_level,
        'parquet_version': parquet_version,
        'field_ids': field_ids,
        'row_group_size': row_group_size,
        'row_group_size_bytes': row_group_size_bytes,
        'row_group_per_file': row_group_per_file,
        'header': header,
        'delimiter': delimiter,
        'quote': quote_char,
        'escape': escape_char,
        'nullstr': null_literal,
        'force_quote': force_quote,
        'prefix': prefix,
        'suffix': suffix,
        'dateformat': date_format,
        'timestampformat': timestamp_format,
        'array': array,
        'per_thread_output': per_thread_output,
        'file_size_bytes': file_size_bytes,
        'partition_by': partition_by,
        'filename_pattern': filename_pattern,
        'file_extension': file_extension,
        'write_partition_columns': write_partition_columns,
        'use_tmp_file': use_tmp_file,
        'overwrite': delete_before_write,
        'overwrite_or_ignore': overwrite,
        'append': append,
        'preserve_order': preserve_order,
    })


@storage.command('update', help='更新存储单元信息')
@argument('name', metavar='NAME')
@option('-r', '--repository', metavar='REPOSITORY', help='所属仓库')
# Postgres；DuckDB；ODPS
@option('--catalog', metavar='CATALOG', help='目录；用于[postgres；duckdb；odps]')
# Postgres；DuckDB；ODPS；SQLite
@option('--schema', metavar='SCHEMA', help='模式；用于[postgres；duckdb；odps；sqlite]')
# Postgres；DuckDB；ODPS；SQLite；BiTable
@option('--table', metavar='TABLE', help='表；用于[postgres；duckdb；odps；sqlite；bitable]')
# BiTable
@option('--document', metavar='DOCUMENT', help='多维表格文档；用于[bitable]')
# File
@option('--file', metavar='FILE', help='文件名；用于[file]')
@option('--format', metavar='FORMAT', type=Choice(['csv', 'parquet', 'json']), help='文件格式；用于[file]')
@option('--compression', metavar='ALGORITHM', type=Choice(['gzip', 'zstd', 'snappy', 'brotli', 'lz4', 'lz4_raw']), help='压缩算法；用于[file]')
@option('--compression-level', metavar='LEVEL', type=INT, help='压缩等级；用于[file][parquet]')
@option('--parquet-version', metavar='VERSION', type=Choice(['V1', 'V2']), help='Parquet文件版本；用于[file][parquet]')
@option('--field-ids', metavar='JSON', type=JSON, help='字段的编号；用于[file][parquet]')
@option('--row-group-size', metavar='SIZE', type=INT, help='最大行数；用于[file][parquet]')
@option('--row-group-size-bytes', metavar='SIZE', type=INT, help='每行组最大字节数；用于[file][parquet]')
@option('--row-group-per-file', metavar='SIZE', type=INT, help='每个文件最大的行组数；用于[file][parquet]')
@option('--header/--no-header', is_flag=True, default=None, help='是否包含字段名；用于[file][csv]')
@option('--delimiter', metavar='DELIMITER', help='字段分隔符；用于[file][csv]')
@option('--quote-char', metavar='CHAR', help='引号字符；用于[file][csv]')
@option('--escape-char', metavar='CHAR', help='转义字符；用于[file][csv]')
@option('--null-literal', metavar='LITERAL', help='空值字面量；用于[file][csv]')
@option('--force-quote', metavar='COLUMN', multiple=True, help='给指定列强制添加引号；用于[file][csv]')
@option('--prefix', metavar='PREFIX', help='前置内容；用于[file][csv]')
@option('--suffix', metavar='SUFFIX', help='后置内容；用于[file][csv]')
@option('--date-format', metavar='FORMAT', help='日期格式；用于[file][csv；json]')
@option('--timestamp-format', metavar='FORMAT', help='时间戳格式；用于[file][csv；json]')
@option('--array/--no-array', is_flag=True, default=None, help='写入JSON数组；用于[file][json]')
@option('--per-thread-output', is_flag=True, default=None, help='是否每个线程写入独立文件；用于[file][thread]')
@option('--file-size-bytes', metavar='BYTES', type=INT, help='每个独立文件最大字节数；用于[file][thread]')
@option('--partition-by', metavar='COLUMN', multiple=True, help='分区列；用于[file][partition]')
@option('--filename-pattern', metavar='PATTERN', help='文件名模式；用于[file][partition]')
@option('--file-extension', metavar='EXTENSION', help='扩展名；用于[file][partition]')
@option('--write-partition-columns', is_flag=True, default=None, help='是否包含分区列；用于[file][partition]')
@option('--use-tmp-file', is_flag=True, default=None, help='若原始文件存在，是否先写入临时文件；用于[file][partition]')
@option('--delete-before-write', is_flag=True, default=None, help='写入之前删除整个目录；用于[file][partition]')
@option('--overwrite/--no-overwrite', is_flag=True, default=None, help='覆盖已存在的文件；用于[file][partition]')
@option('--append/--no-append', is_flag=True, default=None, help='追加到已存在的文件；用于[file][partition]')
@option('--preserve-order', is_flag=True, default=None, help='是否保留原始顺序；用于[file]')
# Others
@help_option('-h', '--help', help='展示帮助信息')
def storage_update(
        name: str,
        repository: str,
        # Postgres；DuckDB；ODPS
        catalog: str,  # 目录；用于[postgres；duckdb；odps]
        # Postgres；DuckDB；ODPS；SQLite
        schema: str,  # 模式；用于[postgres；duckdb；odps；sqlite]
        # Postgres；DuckDB；ODPS；SQLite；BiTable
        table: str,  # 表；用于[postgres；duckdb；odps；sqlite；bitable]
        # BiTable
        document: str,  # 多维表格文档；用于[bitable]
        # File
        file: str,  # 文件名；用于[file]
        format: str,  # 文件格式；用于[file]
        compression: str,  # 压缩算法；用于[file]
        compression_level: int,  # 压缩等级；用于[file][parquet]
        parquet_version: str,  # Parquet文件版本；用于[file][parquet]
        field_ids: dict[str, Any],  # 字段的编号；用于[file][parquet]
        row_group_size: int,  # 最大行数；用于[file][parquet]
        row_group_size_bytes: int,  # 每行组最大字节数；用于[file][parquet]
        row_group_per_file: int,  # 每个文件最大的行组数；用于[file][parquet]
        header: bool,  # 是否包含字段名；用于[file][csv]
        delimiter: str,  # 字段分隔符；用于[file][csv]
        quote_char: str,  # 引号字符；用于[file][csv]
        escape_char: str,  # 转义字符；用于[file][csv]
        null_literal: str,  # 空值字面量；用于[file][csv]
        force_quote: list[str],  # 给指定列强制添加引号；用于[file][csv]
        prefix: str,  # 前置内容；用于[file][csv]
        suffix: str,  # 后置内容；用于[file][csv]
        date_format: str,  # 日期格式；用于[file][csv；json]
        timestamp_format: str,  # 时间戳格式；用于[file][csv；json]
        array: bool,  # 写入JSON数组；用于[file][json]
        per_thread_output: bool,  # 是否每个线程写入独立文件；用于[file][thread]
        file_size_bytes: int,  # 每个独立文件最大字节数；用于[file][thread]
        partition_by: list[str],  # 分区列；用于[file][partition]
        filename_pattern: str,  # 文件名模式；用于[file][partition]
        file_extension: str,  # 扩展名；用于[file][partition]
        write_partition_columns: bool,  # 是否包含分区列；用于[file][partition]
        use_tmp_file: bool,  # 若原始文件存在，是否先写入临时文件；用于[file][partition]
        delete_before_write: bool,  # 写入之前删除整个目录；用于[file][partition]
        overwrite: bool,  # 覆盖已存在的文件；用于[file][partition]
        append: bool,  # 追加到已存在的文件；用于[file][partition]
        preserve_order: bool,  # 是否保留原始顺序；用于[file]
):
    logger.debug(
        'name=%s, repository=%s, catalog=%s, schema=%s, table=%s, document=%s, file=%s, format=%s, compression=%s, compression_level=%s, parquet_version=%s, field_ids=%s, row_group_size=%s, row_group_size_bytes=%s, row_group_per_file=%s, header=%s, delimiter=%s, quote_char=%s, escape_char=%s, null_literal=%s, force_quote=%s, prefix=%s, suffix=%s, date_format=%s, timestamp_format=%s, array=%s, per_thread_output=%s, file_size_bytes=%s, partition_by=%s, filename_pattern=%s, file_extension=%s, write_partition_columns=%s, use_tmp_file=%s, delete_before_write=%s, overwrite=%s, append=%s, preserve_order=%s',
        name, repository, catalog, schema, table, document,
        file, format, compression, compression_level,
        parquet_version, field_ids, row_group_size, row_group_size_bytes, row_group_per_file,
        header, delimiter, quote_char, escape_char, null_literal, force_quote, prefix, suffix,
        date_format, timestamp_format, array,
        per_thread_output, file_size_bytes,
        partition_by, filename_pattern, file_extension, write_partition_columns, use_tmp_file,
        delete_before_write, overwrite, append, preserve_order,
    )
    storage_service.storage_update(repository, name, {
        'catalog': catalog,
        'schema': schema,
        'table': table,
        'document': document,
        'file': file,
        'format': format,
        'compression': compression,
        'compression_level': compression_level,
        'parquet_version': parquet_version,
        'field_ids': field_ids,
        'row_group_size': row_group_size,
        'row_group_size_bytes': row_group_size_bytes,
        'row_group_per_file': row_group_per_file,
        'header': header,
        'delimiter': delimiter,
        'quote': quote_char,
        'escape': escape_char,
        'nullstr': null_literal,
        'force_quote': force_quote,
        'prefix': prefix,
        'suffix': suffix,
        'dateformat': date_format,
        'timestampformat': timestamp_format,
        'array': array,
        'per_thread_output': per_thread_output,
        'file_size_bytes': file_size_bytes,
        'partition_by': partition_by,
        'filename_pattern': filename_pattern,
        'file_extension': file_extension,
        'write_partition_columns': write_partition_columns,
        'use_tmp_file': use_tmp_file,
        'overwrite': delete_before_write,
        'overwrite_or_ignore': overwrite,
        'append': append,
        'preserve_order': preserve_order,
    })


@storage.command('delete', help='删除存储单元；删除迁移')
@argument('name', metavar='NAME')
@option('-r', '--repository', metavar='REPOSITORY', help='所属仓库')
@help_option('-h', '--help', help='展示帮助信息')
def storage_delete(name: str, repository: str):
    logger.debug('name=%s, repository=%s', name, repository)
    storage_service.storage_delete(repository, name)


@storage.command('list', help='列出所有存储单元')
@option('-k', '--kind', type=Choice(RepositoryKind.codes()), help='仓库类型')
@option('-r', '--repository', metavar='REPOSITORY', help='所属仓库')
@help_option('-h', '--help', help='展示帮助信息')
def storage_list(kind: str, repository: str):
    logger.debug('kind=%s, repository=%s', kind, repository)
    table = Table(title='存储单元列表')
    table.add_column('仓库类型', no_wrap=True)
    table.add_column('仓库名称', no_wrap=True)
    table.add_column('存储单元', no_wrap=True)
    table.add_column('关联迁移', justify='right')
    for row in storage_service.storage_list(kind, repository):
        table.add_row(
            row.repository_kind,
            row.repository_code,
            row.code,
            str(row.transformers),
        )

    console = Console()
    console.print(table)



import logging
from typing import Optional, Any

from duckcp.configuration import meta_configuration as metadata
from duckcp.entity.snapshot import Snapshot
from duckcp.entity.storage import Storage
from duckcp.helper.validation import ensure
from duckcp.projection.repository_projection import RepositoryProjection
from duckcp.projection.storage_projection import StorageProjection
from duckcp.repository import RepositoryKind
from duckcp.service import repository_service

logger = logging.getLogger(__name__)


def storage_find(repository_code: str, code: str) -> Optional[Storage]:
    """
    根据编码查找存储单元。
    """
    with metadata.connect() as meta:
        return meta.record('''
          select
            storages.*
          from
            storages
          inner join
            repositories
          on
            storages.repository_id = repositories.id
            and repositories.code = ?
            and storages.code = ?
        ''', repository_code, code, constructor=Storage._make)


def storage_exists(repository_code: str, code: str) -> bool:
    """
    判断编码对应的存储单元是否已存在。
    """
    return storage_find(repository_code, code) is not None


def storage_create(repository_code: str, code: str, properties: dict[str, Any]):
    """
    添加仓库的存储单元。
    """
    logger.debug('repository_code=%s, code=%s, properties=%s', repository_code, code, properties)
    ensure(repository_code is not None, '缺少仓库名称')
    ensure(code is not None, f'仓库({repository_code})缺少存储单元名称')
    properties = {name: value for name, value in properties.items() if bool(value)} if properties else {}
    ensure(bool(properties), f'仓库({repository_code})的存储单元({code})缺少存储参数')

    repository = repository_service.repository_find(repository_code)
    ensure(repository is not None, f'仓库({repository_code})不存在')
    kind = RepositoryKind.of(repository.kind)
    kind.ensure_medium_properties(properties)
    ensure(not storage_exists(repository.code, code), f'仓库({repository_code})的存储单元({code})已存在')

    with metadata.connect() as meta:
        storage = meta.record('''
          insert into storages
            (repository_id, code, properties)
          values
            (?, ?, ?)
          returning *
        ''', repository.id, code, properties, constructor=Storage._make)
        logger.info('创建仓库(%s)的存储单元(%s)', repository.code, code)
        logger.debug('storage=%s', storage)


def storage_update(repository_code: str, code: str, properties: dict[str, Any]):
    """
    更新仓库的存储单元信息。
    """
    logger.debug('repository_code=%s, code=%s, properties=%s', repository_code, code, properties)
    ensure(repository_code is not None, '缺少仓库名称')
    ensure(code is not None, f'仓库({repository_code})缺少存储单元名称')
    properties = {key: value for key, value in properties.items() if value is not None} if properties else {}
    ensure(bool(properties), f'仓库({repository_code})的存储单元({code})缺少更新内容')

    repository = repository_service.repository_find(repository_code)
    ensure(repository is not None, f'仓库({repository_code})不存在')
    storage = storage_find(repository_code, code)
    ensure(storage is not None, f'仓库({repository_code})的存储单元({code})不存在')
    kind = RepositoryKind.of(repository.kind)
    properties = {**storage.properties, **properties}
    kind.ensure_medium_properties(properties)

    with metadata.connect() as meta:
        storage = meta.record('''
          update
            storages
          set
            properties = ?,
            updated_at = datetime(current_timestamp, 'localtime')
          where
            id = ?
          returning *
        ''', {
            key: value
            for key, value in properties.items()
            if bool(value)  # 移除手工强制设为空值的项
        }, storage.id, constructor=Storage._make)
        logger.info('更新仓库(%s)的存储单元(%s)', repository.code, code)
        logger.debug('storage=%s', storage)


def storage_delete(repository_code: str, code: str):
    """
    删除存储单元。
    """
    logger.debug('repository_code=%s, code=%s', repository_code, code)
    ensure(repository_code is not None, '缺少仓库名称')
    ensure(code is not None, f'仓库({repository_code})缺少存储单元名称')

    repository = repository_service.repository_find(repository_code)
    ensure(repository is not None, f'仓库({repository_code})不存在')
    storage = storage_find(repository_code, code)
    ensure(storage is not None, f'仓库({repository_code})的存储单元({code})不存在')

    with metadata.connect() as meta:
        storage = meta.record('''
          delete from storages where id = ? returning *
        ''', storage.id, constructor=Storage._make)
        logger.info('删除仓库(%s)的存储单元(%s)', repository.code, code)
        logger.debug('storage=%s', storage)


def storage_list(
        repository_kind: Optional[str] = None,
        repository_code: Optional[str] = None
) -> list[StorageProjection]:
    """
    列出符合条件的存储单元。
    """
    logger.debug('repository_kind=%s, repository_code=%s', repository_kind, repository_code)

    filters = []
    parameters = []
    if repository_kind is not None:
        RepositoryKind.ensure(repository_kind)
        filters.append('and repositories.kind = ?')
        parameters.append(repository_kind)
    if repository_code is not None:
        filters.append('and repositories.code = ?')
        parameters.append(repository_code)

    with metadata.connect() as meta:
        return meta.records(f'''
          with storages_transformers as (
            select
              target_id as storage_id,
              count(*)  as transformers
            from
              transformers
            group by
              target_id
          )
          select
            repositories.kind as repository_kind,
            repositories.code as repository_code,
            storages.code,
            coalesce(storages_transformers.transformers, 0) as transformers
          from
            storages
          inner join
            repositories
          on
            storages.repository_id = repositories.id
            {' '.join(filters)}
          left join
            storages_transformers
          on
            storages.id = storages_transformers.storage_id
          order by
            repositories.kind,
            repositories.code,
            storages.code
        ''', *parameters, constructor=StorageProjection._make)



from rich.prompt import Prompt


def ensure(condition: bool, message: str = '参数异常'):
    """
    Assert的运行时版本。
    """
    if not condition:
        raise AssertionError(message)


def confirm(prompt: str, default: str = 'n') -> bool:
    return Prompt.ask(prompt, choices=['y', 'n'], default=default).lower() == 'y'



"""
文件系统帮助函数。
"""
import logging
from functools import lru_cache
from os import makedirs, unlink, getcwd, chdir
from os.path import abspath, expanduser, expandvars, exists, join, dirname
from platform import system
from shutil import rmtree, move

logger = logging.getLogger(__name__)


class UnsupportedPlatformError(Exception):
    """
    当前操作系统不支持时，抛出本异常
    """

    def __init__(self, platform: str):
        super().__init__(f'不支持操作系统`{platform}`')


# 文件


def move_file(source: str, target: str):
    """
    移动文件。
    """
    if exists(source):
        logger.info('移动文件(%s)至(%s)', source, target)
        move(source, target)


def remove_file(file: str):
    """
    删除已存在的文件。
    """
    if exists(file):
        logger.info('删除文件(%s)', file)
        unlink(file)


def slurp(file: str) -> str:
    """
    读取文件内容。
    """
    with open(file) as f:
        return f.read()


def spit(file: str, content: str):
    """
    写入文件内容。
    """
    with open(file, 'w') as f:
        return f.write(content)


# 路径

class WorkDirectory:
    """
    临时切换当前工作目录。
    """
    origin: str  # 当前工作目录
    directory: str  # 目标工作目录

    def __init__(self, directory: str):
        self.origin = getcwd()
        self.directory = absolute_path(directory)

    def __enter__(self):
        """
        将工作目录临时切换到指定的目录
        """
        chdir(self.directory)

    def __exit__(self, exception_class, exception, tracebacks):
        """
        将工作目录临时切换回原始目录
        """
        chdir(self.origin)


def absolute_path(path: str) -> str:
    """
    将路径转成绝对路径。
    """
    return abspath(expanduser(expandvars(path)))


def ensure_folder(path: str, parent: bool = False) -> str:
    """
    确保指定的文件夹或父级文件夹存在。
    """
    folder = dirname(path) if parent else path
    if not exists(folder):
        logger.info('创建目录(%s)', folder)
        makedirs(folder)
    return path


def remove_folder(folder: str):
    """
    递归地删除目录。
    """
    if exists(folder):
        logger.info('删除目录(%s)', folder)
        rmtree(folder)


# XDG

@lru_cache()
def xdg(folder: str, *path: str) -> str:
    """
    XDG Base Directory Specification。
    """
    file = join(absolute_path(folder), *path)
    parent = dirname(file)
    if not exists(parent):
        makedirs(parent)
    return file


@lru_cache()
def cache(*path: str) -> str:
    """
    基于缓存目录的路径。
    """
    match system():
        case 'Linux':
            folder = f'~/.cache'
        case 'Darwin':
            folder = f'~/Library/Caches'
        case 'Windows':
            folder = f'%LOCALAPPDATA%\\Temp'
        case _:
            raise UnsupportedPlatformError(system())
    return xdg(folder, *path)


@lru_cache()
def config(*path: str) -> str:
    """
    基于配置目录的路径。
    """
    match system():
        case 'Linux':
            folder = f'~/.config'
        case 'Darwin':
            folder = f'~/Library/Application Support'
        case 'Windows':
            folder = f'%LOCALAPPDATA%'
        case _:
            raise UnsupportedPlatformError(system())
    return xdg(folder, *path)



from typing import NamedTuple


class TransformContext(NamedTuple):
    """
    迁移的上下文。
    """
    source_repository_code: str  # 来源仓库编码
    target_repository_code: str  # 去向仓库编码
    target_storage_code: str  # 目标存储编码



from logging.config import dictConfig
from os import makedirs
from os.path import exists, dirname

import click
from rich.text import Text


def highlighter(text: Text) -> Text:
    """
    自定义的高亮：根据日志级别展示不同的色彩。
    """
    fragments = text.plain.split(' ', 5)
    if len(fragments) == 6:
        (date, time, level, fn, separator, message) = fragments
        if level == 'DEBUG':
            style = 'dim black'
        elif level == 'INFO':
            style = 'cyan'
        elif level == 'WARNING':
            style = 'yellow'
        elif level == 'ERROR':
            style = 'bold red'
        elif level == 'CRITICAL':
            style = 'red on yellow'
        else:
            style = 'black'
        record = f'[dim][cyan]{date} {time}[/cyan] [{style}]{level}[/{style}] {fn} {separator}[/dim] [{style}]{message}[/{style}]'
        return Text.from_markup(record)
    else:
        return text


def enable_logging_configuration(
        file: str | None,
        levels: list[tuple[str, str]],
        message_only: bool,
        verbose: bool,
        quiet: bool
):
    """
    日志输出配置。
    """
    default_level = 'DEBUG' if verbose else 'INFO'
    formatter = 'message' if message_only else 'base'

    if quiet:
        handler = {
            'class': 'logging.NullHandler'
        }
    elif file is None or file == '':  # 不输出文件，则默认输出到控制台
        handler = {
            'class': 'rich.logging.RichHandler',
            'level': default_level,
            'formatter': formatter,
            'highlighter': highlighter,
            'show_time': False,
            'show_level': False,
            'rich_tracebacks': True,
            'tracebacks_suppress': [click],
        }
    else:
        folder = dirname(file)
        if not exists(folder):
            makedirs(folder)
        handler = {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'level': default_level,
            'formatter': formatter,
            'filename': file,
            'when': 'D',
            'backupCount': 30,
        }

    dictConfig({
        'version': 1,
        'formatters': {
            'message': {
                'format': '%(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S',
            },
            'base': {
                'format': '%(asctime)s.%(msecs)03d %(levelname)s %(name)s#%(funcName)s : %(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S',
            },
        },
        'handlers': {
            'handler': handler,
        },
        'root': {
            'level': default_level,
            'propagate': False,
            'handlers': ['handler'],
        },
        'loggers': {
            'duckcp': {
                'level': default_level,
                'propagate': False,
                'handlers': ['handler'],
            },
            '__main__': {
                'level': default_level,
                'propagate': False,
                'handlers': ['handler'],
            },
            **{name: {
                'level': level.upper(),
                'propagate': False,
                'handlers': ['handler'],
            } for name, level in levels}
        },
    })



"""
多维表格接口
"""
import logging
from typing import TypedDict, NotRequired, Any

from duckcp.feishu import OPEN_API, FeiShuError
from duckcp.helper import http
from duckcp.helper.collection import chunk

LIST_FIELDS_API = f'{OPEN_API}/bitable/v1/apps/{{document}}/tables/{{table}}/fields'
LIST_RECORDS_API = f'{OPEN_API}/bitable/v1/apps/{{document}}/tables/{{table}}/records/search'
BATCH_CREATE_API = f'{OPEN_API}/bitable/v1/apps/{{document}}/tables/{{table}}/records/batch_create'
BATCH_DELETE_API = f'{OPEN_API}/bitable/v1/apps/{{document}}/tables/{{table}}/records/batch_delete'
logger = logging.getLogger(__name__)


class Record(TypedDict):
    """
    多维表格单行记录。
    """
    record_id: NotRequired[str]  # 记录编号：创建时可忽略。
    fields: dict[str, Any]  # 字段数据。


class Field(TypedDict):
    """
    多维表格字段。
    """
    field_id: str
    field_name: str
    type: int
    ui_type: str
    is_primary: bool
    is_hidden: bool
    description: str


class Page[T](TypedDict):
    """
    分页查询结果
    """
    has_more: bool
    page_token: str
    total: int
    items: list[T]


class BatchDeleteRecord(TypedDict):
    """
    批量删除结果响应。
    """
    deleted: bool  # 是否删除成功。
    record_id: str  # 记录编号。


class Batch[T](TypedDict):
    """
    批量操作返回的结果。
    """
    records: list[T]  # 记录集。


class Response[T](TypedDict):
    """
    操作结果响应。
    """
    code: int  # 错误编码。
    msg: str  # 错误消息。
    data: T  # 结果。


def list_fields(access_token: str, document: str, table: str) -> list[Field]:
    url = LIST_FIELDS_API.format(document=document, table=table)
    response: Response[Page[Field]] = http.get(url, headers={'Authorization': f'Bearer {access_token}'}, query={'page_size': 100})
    if response['code'] == 0:
        return response['data']['items']
    else:
        raise FeiShuError('获取字段', response.get('msg', ''))


def list_records(access_token: str, document: str, table: str) -> list[Record]:
    """
    分页查询所有记录。
    """
    url = LIST_RECORDS_API.format(document=document, table=table)
    headers = {'Authorization': f'Bearer {access_token}'}
    query: dict[str, Any] = {'page_size': 500}
    records: list[Record] = []
    while True:
        response: Response[Page[Record]] = http.post(url, headers=headers, query=query, params={})
        if response['code'] == 0:
            records.extend(response['data']['items'])
            if response['data']['has_more']:
                query['page_token'] = response['data']['page_token']
            else:
                break
        else:
            raise FeiShuError('获取记录', response.get('msg', ''))
    return records


def batch_create(access_token: str, document: str, table: str, records: list[Record]) -> list[Record]:
    """
    批量创建记录。
    @param access_token: 访问凭证。
    @param document: 多维文档编号。
    @param table: 多维表格编号。
    @param records: 记录编号集合。
    @return: 成功创建的记录。
    """
    url = BATCH_CREATE_API.format(document=document, table=table)
    result: list[Record] = []
    for bucket in chunk(records, 1000):
        response: Response[Batch[Record]] = http.post(url, headers={'Authorization': f'Bearer {access_token}'}, params={'records': bucket})
        if response['code'] == 0:
            result.extend(response['data']['records'])
        else:
            raise FeiShuError('批量创建', response.get('msg', ''))
    logger.debug('rows=%s', len(result))
    return result


def batch_delete(access_token: str, document: str, table: str, records: list[str]) -> list[str]:
    """
    批量删除记录。
    @param access_token: 访问凭证。
    @param document: 多维文档编号。
    @param table: 多维表格编号。
    @param records: 记录编号集合。
    @return: 删除成功的记录编号集合。
    """
    url = BATCH_DELETE_API.format(document=document, table=table)
    result = []
    for bucket in chunk(records, 500):
        response: Response[Batch[BatchDeleteRecord]] = http.post(url, headers={'Authorization': f'Bearer {access_token}'}, params={'records': bucket})
        if response['code'] == 0:
            result.extend([record['record_id'] for record in response['data']['records'] if record['deleted']])
        else:
            raise FeiShuError('批量删除', response.get('msg', ''))
    logger.debug('rows=%s', len(result))
    return result



import logging
import sys
from logging import DEBUG

from duckcp.boot import app, meta_command, repository_command, storage_command, transformer_command, task_command

logger = logging.getLogger(__name__)


def main():
    try:
        app()
    except Exception as e:
        if logger.isEnabledFor(DEBUG):
            logger.exception(e)
        else:
            logger.error('%s', e)
        sys.exit(1)



"""
数据迁移至管系统数据库表，原理如下：
1. 在来源仓库上执行SQL。
2. 根据查询结果生成DELETE语句与INSERT语句。
3. 先执行删除语句清空表。
4. 再执行插入语句新增记录。
"""
import logging

from duckcp.entity.repository import Repository
from duckcp.entity.statement import Statement
from duckcp.entity.storage import Storage
from duckcp.helper.sql import delete_from, insert_into

logger = logging.getLogger(__name__)


def database_transform(statement: Statement, repository: Repository, storage: Storage):
    """
    将数据源迁移到关系型数据库表中。
    """
    catalog = storage.properties.get('catalog')
    schema = storage.properties.get('schema')
    table = storage.properties['table']

    # 通过sqlglot生成SQL，避免字符串转义或SQL注入等问题。
    with repository.connect() as connection:
        with connection.executor() as executor:
            sql = delete_from(catalog, schema, table).sql()
            logger.info('清空表(%s)', sql)
            executor.execute(sql)

            columns, records = statement.execute()
            sql = insert_into(catalog, schema, table, columns).sql()
            logger.info('批量添加数据(%s)', sql)
            executor.batch(sql, records)



"""
序列化与反序列化帮助函数。
"""
from decimal import Decimal
from json import JSONEncoder, dumps, loads
from typing import Any


class JsonEncoder(JSONEncoder):
    """
    序列化JSON的数据：
    - 添加Decimal类型支持。
    - 其他类型数据通过`vars`支持。
    """

    def default(self, value) -> any:
        return float(value) if isinstance(value, Decimal) else vars(value)


def json_encode(data: Any) -> str:
    """
    将数据序列化成JSON字符串：
    - 保留中文
    - 去掉多于空格
    - 支持自定义的类型
    """
    return dumps(data, separators=(',', ':'), ensure_ascii=False, cls=JsonEncoder)


def json_decode(data: str | bytes) -> any:
    """
    将JSON字符串反序列化成对象
    """
    return loads(data)



import logging

from odps import ODPS, dbapi as odps

from duckcp.entity.repository import Repository
from duckcp.helper.validation import ensure

logger = logging.getLogger(__name__)


class OdpsRepository(Repository):
    """
    MaxCompute(ODPS)类型仓库。
    """

    def establish_connection(self) -> odps.Connection:
        """
        创建Odps仓库连接。
        """
        ensure(bool(self.properties), '缺少连接参数')
        ensure(bool(self.properties.get('end_point')), '缺少接入点')
        end_point = self.properties.get('end_point')
        ensure(bool(self.properties.get('project')), '缺少项目名')
        project = self.properties.get('project')
        ensure(bool(self.properties.get('access_key')), '缺少访问编码')
        access_key = self.properties.get('access_key')
        ensure(bool(self.properties.get('access_secret')), '缺少访问密钥')
        access_secret = self.properties.get('access_secret')

        logger.debug('end_point=%s, project=%s, access_key=%s', end_point, project, access_key)
        db = ODPS(access_key, access_secret, project, end_point)
        return odps.connect(db)



import logging
from contextlib import contextmanager
from typing import Any, Optional, Sequence, Iterator

from duckcp.configuration import meta_configuration as metadata
from duckcp.entity.executor import Executor
from duckcp.entity.repository import Repository
from duckcp.helper.validation import ensure
from duckcp.projection.repository_projection import RepositoryProjection
from duckcp.repository import repository_constructor, RepositoryKind

logger = logging.getLogger(__name__)


def repository_find[T: Repository](code: str) -> Optional[T]:
    """
    根据编码查找仓库。
    """
    with metadata.connect() as meta:
        return meta.record('select * from repositories where code = ?', code, constructor=repository_constructor)


def repository_exists(code: str) -> bool:
    """
    判断编码对应的仓库是否已存在。
    """
    return repository_find(code) is not None


def repository_storages(code: str) -> int:
    """
    统计仓库的存储数。
    """
    with metadata.connect() as meta:
        return meta.value('''
          select
            count(*) as tables
          from
            repositories
          inner join
            storages
          on
            repositories.id = storages.repository_id
            and repositories.code = ?
        ''', code)


def repository_transformers(code: str) -> int:
    """
    统计仓库的迁移数。
    """
    with metadata.connect() as meta:
        return meta.value('''
          select
            count(*) as transformers
          from
            repositories
          inner join
            transformers
          on
            repositories.id = transformers.source_id
            and repositories.code = ?
        ''', code)


def repository_create(code: str, kind_code: str, properties: dict[str, Any]):
    """
    创建仓库。
    """
    logger.debug('code=%s, kind_code=%s, properties=%s', code, kind_code, properties)
    ensure(code is not None, '缺少仓库名称')
    ensure(kind_code is not None, f'仓库({code})缺少类型')
    kind = RepositoryKind.of(kind_code)
    ensure(bool(properties), f'仓库({code})缺少连接参数')
    kind.ensure_connection_properties(properties)
    ensure(not repository_exists(code), f'仓库({code})已存在')

    with metadata.connect() as meta:
        repository = meta.record('''
          insert into repositories
            (code, kind, properties)
          values
            (?, ?, ?)
          returning *
        ''', code, kind.code, {
            key: value
            for key, value in properties.items()
            if bool(value)
        }, constructor=repository_constructor)
        logger.info('创建仓库(%s)', code)
        logger.debug('repository=%s', repository)


def repository_update(code: str, kind_code: str, properties: dict[str, Any]):
    """
    更新仓库信息。
    """
    logger.debug('code=%s, kind_code=%s, properties=%s', code, kind_code, properties)
    ensure(code is not None, '缺少仓库名称')
    properties = {key: value for key, value in properties.items() if value is not None} if properties is not None else {}
    ensure(kind_code is not None or bool(properties), f'仓库({code})缺少更新内容')

    repository = repository_find(code)
    ensure(repository is not None, f'仓库({code})不存在')

    kind = RepositoryKind.of(kind_code or repository.kind)
    properties = {**repository.properties, **properties}
    kind.ensure_connection_properties(properties)

    with metadata.connect() as meta:
        repository = meta.record('''
          update
            repositories
          set
            kind = ?,
            properties = ?,
            updated_at = current_timestamp
          where
            code = ?
          returning *
        ''', kind.code, {
            key: value
            for key, value in properties.items()
            if bool(value)  # 移除手工强制设为空值的项
        }, code, constructor=repository_constructor)
        logger.info('更新仓库(%s)', code)
        logger.debug('repository=%s', repository)


def repository_delete(code: str):
    """
    删除仓库。
    - 解绑数据表。
    - 删除迁移。
    """
    logger.debug('code=%s', code)
    ensure(code is not None, '缺少仓库名称')
    ensure(repository_exists(code), f'仓库({code})不存在')
    with metadata.connect() as meta:
        repository = meta.record('''
          delete from repositories where code = ? returning *
        ''', code, constructor=repository_constructor)
        logger.info('删除仓库(%s)', code)
        logger.debug('repository=%s', repository)


def repository_list(kind: Optional[str] = None) -> list[RepositoryProjection]:
    """
    列出符合条件的仓库。
    """
    logger.debug('kind=%s', kind)
    filters = []
    parameters = []
    if kind is not None:
        RepositoryKind.ensure(kind)
        filters.append('where repositories.kind = ?')
        parameters.append(kind)

    with metadata.connect() as meta:
        return meta.records(f'''
          with repositories_transformers as (
            select
              repositories.id as repository_id,
              count(*) as transformers
            from
              repositories
            inner join
              transformers
            on
              repositories.id = transformers.source_id
            group by
              repositories.id
          ), repositories_storages as (
            select
              repositories.id as repository_id,
              count(*) as storages
            from
              repositories
            inner join
              storages
            on
              repositories.id = storages.repository_id
            group by
              repositories.id
          )
          select
            repositories.kind,
            repositories.code,
            coalesce(repositories_storages.storages, 0) as storages,
            coalesce(repositories_transformers.transformers, 0) as transformers
          from
            repositories
          left join
            repositories_transformers
          on
            repositories.id = repositories_transformers.repository_id
          left join
            repositories_storages
          on
            repositories.id = repositories_storages.repository_id
          {' '.join(filters)}
          order by
            repositories.kind,
            repositories.code
        ''', *parameters, constructor=RepositoryProjection._make)


@contextmanager
def repository_connect(code: str) -> Iterator[Executor]:
    """
    连接指定的仓库。
    """
    logger.debug('code=%s', code)
    repository = repository_find(code)
    ensure(repository is not None, f'仓库({code})不存在')

    with repository.connect() as connection:
        with connection.executor() as executor:
            yield executor


def repository_execute(code: str, sql: str) -> tuple[list[str], list[Sequence[Any]]]:
    """
    在仓库上执行查询脚本，并返回查询结果。
    """
    logger.debug('code=%s, sql=%s', code, sql)
    with repository_connect(code) as executor:
        return executor.execute(sql)



"""
数据迁移至DuckDB数据库表，原理如下：
1. 在来源仓库上执行SQL，并将查询结果封装成DataFrame。
2. 将DataFrame映射成DuckDB的只读视图。
3. 执行`create ore replace table ... from ...`替换目标表内的数据。
"""
import logging

from duckcp.entity.statement import Statement
from duckcp.entity.storage import Storage
from duckcp.helper.sql import create_or_replace_table
from duckcp.repository.duckdb_repository import DuckDBRepository

logger = logging.getLogger(__name__)


def duckdb_transform(statement: Statement, repository: DuckDBRepository, storage: Storage):
    """
    将数据源迁移到DuckDB数据库表中。
    """
    catalog = storage.properties.get('catalog')
    schema = storage.properties.get('schema')
    table = storage.properties['table']

    with repository.establish_connection() as connection:
        with connection.cursor() as cursor:
            data = statement()
            cursor.execute(f' set global pandas_analyze_sample = {len(data)} ')
            cursor.register(storage.code, data)  # 表的全名是`temp.main.<table>`
            # 通过sqlglot生成SQL，避免字符串转义或SQL注入等问题。
            ast = create_or_replace_table(catalog, schema, table, storage.code)
            sql = ast.sql(dialect='duckdb')
            logger.debug('sql=%s', sql)
            cursor.execute(sql)



from typing import Callable

from duckcp.entity.repository import Repository
from duckcp.entity.statement import Statement
from duckcp.entity.storage import Storage

# 定义迁移函数接口
type Transform[T: Repository] = Callable[[Statement, T, Storage], None]



import logging

import psycopg2

from duckcp.entity.repository import Repository
from duckcp.helper.validation import ensure

logger = logging.getLogger(__name__)


class PostgresRepository(Repository):
    """
    Postgres类型仓库。
    """

    def establish_connection(self) -> psycopg2.extensions.connection:
        """
        创建Postgres连接。
        """
        ensure(bool(self.properties), '缺少连接参数')
        host = self.properties.get('host')
        port = self.properties.get('port')
        ensure(bool(self.properties.get('database')), '缺少数据库名称')
        database = self.properties.get('database')
        username = self.properties.get('username')
        password = self.properties.get('password')
        logger.debug('host=%s, port=%s, database=%s, username=%s', host, port, database, username)

        return psycopg2.connect(
            host=host,
            port=port,
            database=database,
            user=username,
            password=password
        )



"""
元信息数据库管理服务。
"""
import logging
from importlib.resources import path
from os import unlink, chmod
from os.path import exists

from duckcp import migration
from duckcp.configuration import meta_configuration as metadata, Configuration

logger = logging.getLogger(__name__)


def meta_create():
    """
    创建元信息数据库。
    """
    if not exists(Configuration.file):
        logger.info('配置文件(%s)初始化', Configuration.file)
        with metadata.connect() as meta:
            with path(migration) as folder:
                for script in sorted(folder.glob('**/*.sql')):
                    logger.info('执行脚本(%s)', script.name)
                    meta.execute(script.read_text(encoding='utf-8'))
        chmod(Configuration.file, 0o600)  # 配置文件里包含部分敏感信息，因此只允许当前用户访问
    else:
        logger.warning('配置文件(%s)已存在', Configuration.file)


def meta_delete():
    """
    删除元信息数据库。
    """
    if exists(Configuration.file):
        logger.info('删除配置文件(%s)', Configuration.file)
        unlink(Configuration.file)
    else:
        logger.info('配置文件(%s)不存在；忽略删除操作', Configuration.file)



"""
摘要算法帮助函数
"""
import hashlib
from typing import Callable


# bytes版本

def hash_encode_bytes(algorithm: Callable, data: bytes) -> bytes:
    """
    通过Hash算法编码明文数据。
    :param algorithm: hashlib包里的算法。
    :param data: 明文字节流。
    :return: 密文字节流。
    """
    encoder = algorithm()
    encoder.update(data)
    return encoder.digest()


def md5_bytes(data: bytes) -> bytes:
    """
    计算明文数据的MD5摘要。
    :param data: 明文字节流。
    :return: MD5字节流。
    """
    return hash_encode_bytes(hashlib.md5, data)


def sha256_bytes(data: bytes) -> bytes:
    """
    计算明文数据的SHA256摘要。
    :param data: 明文字节流。
    :return: SHA256字节流。
    """
    return hash_encode_bytes(hashlib.sha256, data)


def sha512_bytes(data: bytes) -> bytes:
    """
    计算明文数据的SHA512摘要。
    :param data: 明文字节流。
    :return: SHA512字节流。
    """
    return hash_encode_bytes(hashlib.sha512, data)


# str 版本

def hash_encode(algorithm: Callable, data: str) -> str:
    """
    通过Hash算法编码明文数据。
    :param algorithm: hashlib包里的算法。
    :param data: 明文字符串。
    :return: 密文字符串。
    """
    return hash_encode_bytes(algorithm, data.encode('utf-8')).hex()


def md5(data: str) -> str:
    """
    计算明文数据的MD5摘要。
    :param data: 明文字符串。
    :return: MD5字符串。
    """
    return hash_encode(hashlib.md5, data)


def sha256(data: str) -> str:
    """
    计算明文数据的SHA256摘要。
    :param data: 明文字符串。
    :return: SHA256字符串。
    """
    return hash_encode(hashlib.sha256, data)


def sha512(data: str) -> str:
    """
    计算明文数据的SHA512摘要。
    :param data: 明文字符串。
    :return: SHA512字符串。
    """
    return hash_encode(hashlib.sha512, data)



import logging

import duckdb
from duckdb.duckdb import DuckDBPyConnection

from duckcp.entity.repository import Repository

logger = logging.getLogger(__name__)


class DuckDBRepository(Repository):
    """
    DuckDB类型仓库。
    """

    def establish_connection(self) -> DuckDBPyConnection:
        """
        创建DuckDB数据库连接。
        """
        file = self.properties['file'] if self.properties and 'file' in self.properties else ':memory:'
        logger.debug('file=%s', file)
        return duckdb.connect(file)



from typing import Self, Sequence, Any, Optional

from pandas import DataFrame

from duckcp.entity.executor import Executor
from duckcp.typing.record_constructor_protocol import RecordConstructorProtocol


class Statement:
    """
    语句：预置SQL语句的执行器。
    """
    executor: Executor
    sql: str

    def __init__(self, executor: Executor, sql: str):
        self.executor = executor
        self.sql = sql

    def __enter__(self) -> Self:
        """
        生命周期开始。
        """
        return self

    def __exit__(self, exception_class, exception, traceback):
        """
        生命周期结束。
        """
        self.close()

    def close(self):
        """
        关闭本次会话/游标。
        """
        self.executor.close()

    def batch(self, parameters: list[Sequence[Any]]):
        """
        批量执行语句，无返回结果。
        """
        self.executor.batch(self.sql, parameters)

    def execute(self, *parameters: Any) -> tuple[list[str], list[Sequence[Any]]]:
        """
        执行单条语句，返回原始的结果。
        - 头信息：列名和类型。
        - 记录：原始数据。
        """
        return self.executor.execute(self.sql, *parameters)

    def __call__(self, *parameters: Any) -> DataFrame:
        """
        执行查询语句，返回DataFrame结构。
        """
        return self.executor(self.sql, *parameters)

    def records[T: tuple[Any, ...]](self, *parameters: Any, constructor: RecordConstructorProtocol[T] = None) -> list[T]:
        """
        执行查询语句，返回NamedTuple形式的记录数据。
        """
        return self.executor.records(self.sql, *parameters, constructor=constructor)

    def record[T: tuple[Any, ...]](self, *parameters: Any, constructor: RecordConstructorProtocol[T] = None) -> Optional[T]:
        """
        执行查询语句，并返回第一行NamedTuple形式的记录数据。
        """
        return self.executor.record(self.sql, *parameters, constructor=constructor)

    def value(self, *parameters: Any) -> Optional[Any]:
        """
        执行查询语句，并返回第一行第一列数据。
        """
        return self.executor.value(self.sql, *parameters)

    def values(self, *parameters: Any) -> list[Any]:
        """
        执行查询语句，并返回所有行的第一列数据。
        """
        return self.executor.values(self.sql, *parameters)

    def all(self, *parameters: Any) -> list[dict[str, Any]]:
        """
        执行查询语句，返回字典结构的记录。
        """
        return self.executor.all(self.sql, *parameters)

    def one(self, *parameters: Any) -> dict[str, Any]:
        """
        执行查询语句，返回字典结构的第一行记录。
        """
        return self.executor.one(self.sql, *parameters)



"""
定时任务调度服务。
"""
import logging
from typing import Optional

from duckcp.configuration import meta_configuration as metadata
from duckcp.entity.task import Task
from duckcp.entity.task_transformer import TaskTransformer
from duckcp.helper.validation import ensure
from duckcp.projection.task_projection import TaskProjection
from duckcp.projection.task_transformer_projection import TaskTransformerProjection
from duckcp.service import transformer_service

logger = logging.getLogger(__name__)


def task_find(code: str) -> Optional[Task]:
    """
    根据编码查找任务。
    """
    with metadata.connect() as meta:
        return meta.record('select * from tasks where code = ?', code, constructor=Task._make)


def task_exists(code: str) -> bool:
    """
    判断编码对应的任务是否已存在。
    """
    return task_find(code) is not None


def task_transformer_find(task_id: int, transformer_id: int) -> TaskTransformer:
    """
    判断任务与迁移是否已绑定。
    """
    with metadata.connect() as meta:
        return meta.record('select * from tasks_transformers where task_id = ? and transformer_id = ?', task_id, transformer_id, constructor=TaskTransformer._make)


def task_transformers(task_id: int) -> int:
    """
    任务关联的迁移数。
    """
    with metadata.connect() as meta:
        return meta.value('select count(*) as transformers from tasks_transformers where task_id = ?', task_id)


def task_create(code: str):
    """
    创建任务。
    """
    logger.debug('code=%s', code)
    ensure(code is not None, '缺少任务名称')
    ensure(not task_exists(code), f'任务({code})已存在')
    with metadata.connect() as meta:
        task = meta.record('''
          insert into tasks (code) values (?) returning *
        ''', code, constructor=Task._make)
        logger.info('创建任务(%s)', code)
        logger.debug('task=%s', task)


def task_delete(code: str):
    """
    删除任务。
    """
    logger.debug('code=%s', code)
    ensure(code is not None, '缺少任务名称')
    ensure(task_exists(code), f'任务({code})不存在')
    with metadata.connect() as meta:
        task = meta.record('''
          delete from tasks where code = ? returning *
        ''', code, constructor=Task._make)
        logger.info('删除任务(%s)', code)
        logger.debug('task=%s', task)


def task_list() -> list[TaskProjection]:
    """
    罗列所有任务。
    """
    with metadata.connect() as meta:
        return meta.records('''
          with task_transformers as (
            select
              task_id,
              count(*) as transformers
            from
              tasks_transformers
            group by
              task_id
          )
          select
            tasks.code,
            coalesce(task_transformers.transformers, 0) as transformers
          from
            tasks
          left join
            task_transformers
          on
            tasks.id = task_transformers.task_id
          group by
            tasks.code
          order by
            tasks.code
        ''', constructor=TaskProjection._make)


def task_execute(code: str):
    """
    执行迁移任务。
    """
    logger.debug('code=%s', code)
    ensure(task_exists(code), f'任务({code})不存在')
    with metadata.connect() as meta:
        for transformer_code in meta.values('''
          select
            transformers.code
          from
            tasks
          inner join
            tasks_transformers
          on
            tasks.id = tasks_transformers.task_id
            and tasks.code = ?
          inner join
            transformers
          on
            tasks_transformers.transformer_id = transformers.id
          order by
            tasks_transformers.sort
        ''', code):
            transformer_service.transformer_execute(transformer_code)


def task_bind(code: str, transformer_code: str, sort: int):
    """
    迁移与任务绑定。
    """
    logger.debug('code=%s, transformer_code=%s', code, transformer_code)
    task = task_find(code)
    ensure(task is not None, f'任务({code})不存在')
    transformer = transformer_service.transformer_find(transformer_code)
    ensure(transformer is not None, f'迁移({transformer_code})不存在')
    ensure(task_transformer_find(task.id, transformer.id) is None, f'任务({code})与迁移({transformer_code})已绑定')
    with metadata.connect() as meta:
        transformers = task_transformers(task.id)
        if sort is not None and 0 < sort <= transformers:
            meta.execute('''
              update
                tasks_transformers
              set
                sort = sort + 1,
                updated_at = datetime(current_timestamp, 'localtime')
              where
                task_id = ?
                and sort >= ?
            ''', task.id, sort)
            logger.info('后移序号(%s-%s)的迁移', sort, transformers)
        else:
            sort = transformers + 1
        task_transformer = meta.record('''
          insert into tasks_transformers
            (task_id, transformer_id, sort)
          values
            (?, ?, ?)
          returning *
        ''', task.id, transformer.id, sort)
        logger.info('绑定任务(%s)与迁移(%s)', code, transformer_code)
        logger.debug('task_transformer=%s', task_transformer)


def task_unbind(code: str, transformer_code: str):
    """
    迁移与任务解绑。
    """
    logger.debug('code=%s, transformer_code=%s', code, transformer_code)
    task = task_find(code)
    ensure(task is not None, f'任务({code})不存在')
    transformer = transformer_service.transformer_find(transformer_code)
    ensure(transformer is not None, f'迁移({transformer_code})不存在')
    task_transformer = task_transformer_find(task.id, transformer.id)
    ensure(task_transformer is not None, f'任务({code})与迁移({transformer_code})未绑定')
    with metadata.connect() as meta:
        transformers = task_transformers(task.id)
        if task_transformer.sort < transformers:
            meta.execute('''
              update
                tasks_transformers
              set
                sort = sort - 1,
                updated_at = datetime(current_timestamp, 'localtime')
              where
                task_id = ?
                and sort > ?
            ''', task.id, task_transformer.sort)
            logger.info('前移序号(%s-%s)的迁移', task_transformer.sort + 1, transformers)
        task_transformer = meta.record('''
          delete from tasks_transformers where task_id = ? and transformer_id = ? returning *
        ''', task.id, transformer.id)
        logger.info('解绑任务(%s)与迁移(%s)', code, transformer_code)
        logger.debug('task_transformer=%s', task_transformer)


def task_transformer_list(code: str) -> list[TaskTransformerProjection]:
    """
    列出任务内所有迁移。
    """
    logger.debug('code=%s', code)
    filters = []
    parameters = []
    if code is not None:
        ensure(task_exists(code), f'任务({code})不存在')
        filters.append('and tasks.code = ?')
        parameters.append(code)

    with metadata.connect() as meta:
        return meta.records(f'''
          select
            tasks.code as task_code,
            tasks_transformers.sort,
            transformers.code as transformer_code,
            sources.kind as source_repository_kind,
            sources.code as source_repository_code,
            targets.kind as target_repository_kind,
            targets.code as target_repository_code,
            storages.code as target_storage_code,
            transformers.script_file
          from
            tasks
          inner join
            tasks_transformers
          on
            tasks.id = tasks_transformers.task_id
            {' '.join(filters)}
          inner join
            transformers
          on
            tasks_transformers.transformer_id = transformers.id
          inner join
            repositories as sources
          on
            transformers.source_id = sources.id
          inner join
            storages
          on
            transformers.target_id = storages.id
          inner join
            repositories as targets
          on
            storages.repository_id = targets.id
          order by
            tasks.code,
            tasks_transformers.sort
        ''', *parameters, constructor=TaskTransformerProjection._make)



"""
查询结果返回记录的构造函数。
"""
from typing import Protocol, Any, Sequence


class RecordConstructorProtocol[T: tuple[Any, ...]](Protocol):
    """
    一个抽象基类，含一个抽象方法 __call__，该方法用于构造所需的记录类型。
    """

    def __call__(self, record: Sequence[Any]) -> T:
        ...



"""
访问凭证的刷新函数
"""
from datetime import datetime
from typing import Callable

from duckcp.typing.authentication_token_type import AuthenticationToken

type CredentialRefresher = Callable[[AuthenticationToken], tuple[str, datetime]]



import logging
from os.path import exists

import click
from click import help_option, option, argument, Choice
from rich.console import Console
from rich.table import Table

from duckcp.boot import app
from duckcp.helper.fs import slurp, absolute_path
from duckcp.helper.validation import ensure, confirm
from duckcp.repository import RepositoryKind
from duckcp.service import repository_service

logger = logging.getLogger(__name__)


@app.group(help='管理仓库')
@help_option('-h', '--help', help='展示帮助信息')
def repository():
    pass


@repository.command('create', help='创建仓库')
@argument('name', metavar='NAME')
@option('-k', '--kind', type=Choice(RepositoryKind.codes()), required=True, help='仓库类型')
# Postgres
@option('--host', metavar='HOST', help='主机；用于[postgres]')
@option('--port', type=click.INT, metavar='PORT', help='端口；用于[postgres]')
@option('--database', metavar='DATABASE', help='数据库名；用于[postgres]')
@option('--username', metavar='USERNAME', help='登入用户；用于[postgres]')
@option('--password', metavar='PASSWORD', help='登入密码；用于[postgres]')
# ODPS
@option('--end-point', metavar='END-POINT', help='地址；用于[odps]')
@option('--project', metavar='PROJECT', help='项目；用于[odps]')
# ODPS；BiTable
@option('--access-key', metavar='KEY', help='凭证编码；用于[odps；bitable]')
@option('--access-secret', metavar='SECRET', help='凭证密钥；用于[odps；bitable]')
# DuckDB; SQLite
@option('--file', metavar='FILE', help='文件；用于[duckdb；sqlite]')
# File
@option('--folder', metavar='FOLDER', help='目录；用于[file]')
# Others
@help_option('-h', '--help', help='展示帮助信息')
def repository_create(
        name: str,
        kind: str,
        # Postgres
        host: str,
        port: int,
        database: str,
        username: str,
        password: str,
        # ODPS; BiTable
        end_point: str,
        project: str,
        access_key: str,
        access_secret: str,
        # DuckDB; SQLite
        file: str,
        # File
        folder: str,
):
    logger.debug(
        'name=%s, kind=%s, host=%s, port=%s, database=%s, username=%s, end_point=%s, project=%s, access_key=%s, file=%s, folder=%s',
        name, kind,
        host, port, database, username,
        end_point, project, access_key,
        file, folder,
    )
    repository_service.repository_create(name, kind, {
        'host': host or None,
        'port': port or None,
        'database': database or None,
        'username': username or None,
        'password': password or None,
        'end_point': end_point or None,
        'project': project or None,
        'access_key': access_key or None,
        'access_secret': access_secret or None,
        'file': absolute_path(file) if file else None,
        'folder': absolute_path(folder) if folder else None,
    })


@repository.command('update', help='更新仓库信息')
@argument('name', metavar='NAME')
@option('-k', '--kind', type=Choice(RepositoryKind.codes()), help='仓库类型')
# Postgres
@option('--host', metavar='HOST', help='主机；用于[postgres]')
@option('--port', type=click.INT, metavar='PORT', help='端口；用于[postgres]')
@option('--database', metavar='DATABASE', help='数据库名；用于[postgres]')
@option('--username', metavar='USERNAME', help='登入用户；用于[postgres]')
@option('--password', metavar='PASSWORD', help='登入密码；用于[postgres]')
# ODPS
@option('--end-point', metavar='END-POINT', help='地址；用于[odps]')
@option('--project', metavar='PROJECT', help='项目；用于[odps]')
# ODPS；BiTable
@option('--access-key', metavar='KEY', help='凭证编码；用于[odps；bitable]')
@option('--access-secret', metavar='SECRET', help='凭证密钥；用于[odps；bitable]')
# DuckDB; SQLite
@option('--file', metavar='FILE', help='文件；用于[duckdb；sqlite]')
# File
@option('--folder', metavar='FOLDER', help='目录；用于[file]')
# Others
@help_option('-h', '--help', help='展示帮助信息')
def repository_update(
        name: str,
        kind: str,
        # Postgres
        host: str,
        port: int,
        database: str,
        username: str,
        password: str,
        # ODPS；BiTable
        end_point: str,
        project: str,
        access_key: str,
        access_secret: str,
        # DuckDB; SQLite
        file: str,
        # File
        folder: str,
):
    logger.debug(
        'name=%s, kind=%s, host=%s, port=%s, database=%s, username=%s, end_point=%s, project=%s, access_key=%s, file=%s, folder=%s',
        name, kind,
        host, port, database, username,
        end_point, project, access_key,
        file, folder,
    )
    repository_service.repository_update(name, kind, {
        'host': host,
        'port': port,
        'database': database,
        'username': username,
        'password': password,
        'end_point': end_point,
        'project': project,
        'access_key': access_key,
        'access_secret': access_secret,
        'file': absolute_path(file) if file else file,
        'folder': absolute_path(folder) if folder else file,
    })


@repository.command('delete', help='注销仓库；删除迁移')
@argument('name', metavar='NAME')
@option('-y', '--yes', is_flag=True, help='自动同意')
@help_option('-h', '--help', help='展示帮助信息')
def repository_delete(name: str, yes: bool):
    logger.debug('name=%s, yes=%s', name, yes)
    transformers = repository_service.repository_transformers(name)
    if transformers > 0 and not yes and not confirm(f'仓库({name})还有{transformers}个关联迁移，确认删除？'):
        return
    tables = repository_service.repository_storages(name)
    if tables > 0 and not yes and not confirm(f'仓库({name})还有{tables}个关联表，确认删除？'):
        return
    repository_service.repository_delete(name)


@repository.command('list', help='列出所有仓库')
@option('-k', '--kind', type=Choice(RepositoryKind.codes()), help='仓库类型')
@help_option('-h', '--help', help='展示帮助信息')
def repository_list(kind: str):
    table = Table(title='仓库列表')
    table.add_column('类型', no_wrap=True)
    table.add_column('名称', no_wrap=True)
    table.add_column('关联存储', justify='right')
    table.add_column('关联迁移', justify='right')
    for row in repository_service.repository_list(kind):
        table.add_row(
            row.kind,
            row.code,
            str(row.storages),
            str(row.transformers),
        )

    console = Console()
    console.print(table)


@repository.command('execute', help='执行脚本')
@argument('name', metavar='NAME')
@option('-f', '--script', metavar='FILE', help='查询数据的SQL脚本文件')
@option('-e', '--expression', metavar='SQL', help='查询数据的SQL')
@help_option('-h', '--help', help='展示帮助信息')
def repository_execute(name: str, script: str, expression: str):
    logger.debug('name=%s, script=%s, expression=%s', name, script, expression)
    if script is None and expression is None:
        raise ValueError('请指定查询SQL的脚本文件或查询语句')
    if script is not None and expression is None:
        ensure(exists(script), f'脚本文件({script})不存在')
        expression = slurp(script)

    table = Table(title=name)
    columns, records = repository_service.repository_execute(name, expression)

    for column in columns:
        table.add_column(column)

    for record in records:
        table.add_row(*map(str, record))

    console = Console()
    console.print(table)



from typing import NamedTuple


class TransformerProjection(NamedTuple):
    code: str  # 编码
    source_repository_kind: str  # 来源仓库类型
    source_repository_code: str  # 来源仓库编码
    target_repository_kind: str  # 目标仓库类型
    target_repository_code: str  # 目标仓库编码
    target_storage_code: str  # 目标存储编码
    script_file: str  # 迁移脚本
    tasks: int  # 关联任务数量



from datetime import datetime
from typing import NamedTuple, Any


class Storage(NamedTuple):
    id: int
    repository_id: int  # 所属仓库
    code: str  # 编码
    properties: dict[str, Any]  # 介质信息
    created_at: datetime
    updated_at: datetime



import logging
from collections.abc import Sequence
from enum import Enum
from typing import Any, NamedTuple

from duckcp.repository.bitable_repository import BiTableRepository
from duckcp.repository.duckdb_repository import DuckDBRepository
from duckcp.repository.file_repository import FileRepository
from duckcp.repository.odps_repository import OdpsRepository
from duckcp.repository.postgres_repository import PostgresRepository
from duckcp.repository.sqlite_repository import SqliteRepository
from duckcp.transform.bitable_transform import bitable_transform
from duckcp.transform.database_transform import database_transform
from duckcp.transform.duckdb_transform import duckdb_transform
from duckcp.transform.file_transform import file_transform
from duckcp.typing.transform_type import Transform

logger = logging.getLogger(__name__)


class RepositoryKind(Enum):
    """
    仓库类型。用于管理不同类型仓库的选项.
    """
    Postgres = (
        'postgres',
        PostgresRepository,
        ['database'],
        ['table'],
        database_transform,
    )
    Odps = (
        'odps',
        OdpsRepository,
        ['end_point', 'project', 'access_key', 'access_secret'],
        ['table'],
        database_transform,
    )
    BiTable = (
        'bitable',
        BiTableRepository,
        ['access_key', 'access_secret'],
        ['document', 'table'],
        bitable_transform,
    )
    DuckDB = (
        'duckdb',
        DuckDBRepository,
        ['file'],
        ['table'],
        duckdb_transform,
    )
    Sqlite = (
        'sqlite',
        SqliteRepository,
        ['file'],
        ['table'],
        database_transform,
    )
    File = (
        'file',
        FileRepository,
        ['folder'],
        ['file'],
        file_transform,
    )

    @staticmethod
    def codes() -> list[str]:
        """
        仓库类型的编码：用于交互选择仓库类型。
        """
        return [kind.code for kind in RepositoryKind]

    @staticmethod
    def of(code: str) -> 'RepositoryKind':
        """
        根据编码获取仓库类型。
        """
        for kind in RepositoryKind:
            if kind.code == code:
                return kind
        else:
            raise ValueError(f'仓库类型({code})不支持')

    @staticmethod
    def ensure(code: str):
        """
        确认支持仓库编码对应的仓库类型。
        """
        RepositoryKind.of(code)

    @property
    def code(self) -> str:
        """
        当前仓库类型的唯一编码。
        """
        return self.value[0]

    @property
    def repository(self) -> type[NamedTuple]:
        """
        当前类型的仓库。
        """
        return self.value[1]

    @property
    def required_connection_options(self) -> list[str]:
        """
        当前类型仓库必要的连接选项。
        """
        return self.value[2]

    def ensure_connection_properties(self, properties: dict[str, Any]):
        """
        确保仓库类型需要的连接选项有提供。
        """
        for name in self.required_connection_options:
            option = f'--{name.replace("_", "-")}'
            if not bool(properties.get(name)):
                raise AssertionError(f'{self.code}类型仓库缺少`{option}`')

    @property
    def required_medium_options(self) -> list[str]:
        """
        当前类型仓库的存储单元中必要的介质选项。
        """
        return self.value[3]

    def ensure_medium_properties(self, properties: dict[str, Any]):
        """
        确保仓库类型的存储单元需要的介质选项有提供。
        """
        for name in self.required_medium_options:
            option = f'--{name.replace("_", "-")}'
            if not bool(properties.get(name)):
                raise AssertionError(f'{self.code}类型仓库的存储缺少`{option}`')

    @property
    def transform(self) -> Transform:
        """
        当前类型仓库的迁移函数。
        """
        return self.value[4]


def repository_constructor[T: tuple](record: Sequence[Any]) -> T:
    """
    根据记录的仓库类型，创建对应类型的仓库。
    """
    logger.debug('record=%s', record)
    kind = RepositoryKind.of(record[1])
    repository_class = kind.repository
    return repository_class._make(record)



from json import JSONDecodeError

from click import ParamType

from duckcp.helper.serialization import json_decode


class JSONParamType(ParamType):
    """
    JSON类型。
    """
    name = 'json'

    def convert(self, value, parameter, context):
        try:
            return json_decode(value)
        except JSONDecodeError:
            self.fail(f"'{value}'不是有效的JSON格式", parameter, context)


JSON = JSONParamType()



import logging
import sqlite3
from typing import Any

from duckcp.entity.repository import Repository
from duckcp.helper.serialization import json_encode, json_decode

logger = logging.getLogger(__name__)


def to_bool(value: bytes) -> bool:
    """
    将SQLite返回的Bytes类型布尔值转成Python的布尔值。
    """
    return bool(int(value)) if value is not None else None


def to_value(value: bytes) -> Any:
    """
    将SQLite的JSON转成Python的值。
    """
    return json_decode(value) if value is not None else None


def to_json(value: Any) -> str:
    """
    将Python的值转成SQLite的JSON。
    """
    return json_encode(value) if value is not None else None


sqlite3.register_converter('boolean', to_bool)
sqlite3.register_converter('bool', to_bool)
sqlite3.register_converter('jsonb', to_value)
sqlite3.register_converter('json', to_value)
sqlite3.register_adapter(dict, to_json)
sqlite3.register_adapter(list, to_json)


class SqliteRepository(Repository):
    """
    Sqlite类型仓库。
    """

    def establish_connection(self) -> sqlite3.Connection:
        """
        创建Sqlite连接。
        """
        file = self.properties['file'] if self.properties and 'file' in self.properties else ':memory:'
        logger.debug('file=%s', file)
        connection = sqlite3.connect(file, detect_types=sqlite3.PARSE_DECLTYPES, autocommit=True)
        connection.execute('PRAGMA foreign_keys=ON')  # 启用on delete cascade
        return connection



from datetime import datetime
from typing import NamedTuple


class TaskTransformer(NamedTuple):
    task_id: int  # 所属任务
    transformer_id: int  # 所属迁移
    sort: int  # 排序
    created_at: datetime
    updated_at: datetime



from typing import Protocol

from duckcp.typing.cursor_protocol import CursorProtocol


class ConnectionProtocol(Protocol):
    """
    一个抽象基类，定义数据库连接协议。
    """

    def close(self):
        """
        断开连接。
        """
        ...

    def cursor(self) -> CursorProtocol:
        """
        创建新的游标对象。
        """
        ...



import logging
from os.path import exists
from typing import Optional

from duckcp.configuration import meta_configuration as metadata
from duckcp.entity.transform_context import TransformContext
from duckcp.entity.transformer import Transformer
from duckcp.helper.fs import absolute_path, slurp
from duckcp.helper.validation import ensure
from duckcp.projection.transformer_projection import TransformerProjection
from duckcp.repository import RepositoryKind
from duckcp.service import repository_service, storage_service

logger = logging.getLogger(__name__)


def transformer_find(code: str) -> Optional[Transformer]:
    """
    根据编码查找迁移。
    """
    with metadata.connect() as meta:
        return meta.record('select * from transformers where code = ?', code, constructor=Transformer._make)


def transformer_exists(code: str) -> bool:
    """
    判断编码对应的迁移是否已存在。
    """
    return transformer_find(code) is not None


def transformer_create(
        code: str,
        source_repository_code: str,
        target_repository_code: str,
        target_storage_code: str,
        script_file: str,
):
    """
    添加迁移。
    """
    logger.debug(
        'code=%s, source_repository_code=%s, target_repository_code=%s, target_storage_code=%s, script_file=%s',
        code, source_repository_code,
        target_repository_code, target_storage_code,
        script_file
    )
    ensure(code is not None, '缺少迁移名称')
    ensure(source_repository_code is not None, f'迁移({code})缺少来源仓库名称')
    ensure(target_repository_code is not None, f'迁移({code})缺少目标仓库名称')
    ensure(target_storage_code is not None, f'迁移({code})缺少目标存储单元')
    ensure(script_file is not None, f'迁移({code})缺少迁移脚本')

    ensure(not transformer_exists(code), f'迁移({code})已存在')
    repository = repository_service.repository_find(source_repository_code)
    ensure(repository is not None, f'来源仓库({source_repository_code})不存在')
    storage = storage_service.storage_find(target_repository_code, target_storage_code)
    ensure(storage is not None, f'目标仓库({target_repository_code})的存储单元({target_storage_code})不存在')
    script_file = absolute_path(script_file)

    with metadata.connect() as meta:
        transformer = meta.record('''
          insert into transformers
            (code, source_id, target_id, script_file)
          values
            (?, ?, ?, ?)
          returning *
        ''', code, repository.id, storage.id, script_file, constructor=Transformer._make)
        logger.info('创建迁移(%s)', code)
        logger.debug('transformer=%s', transformer)


def transformer_update(
        code: str,
        source_repository_code: str,
        target_repository_code: str,
        target_storage_code: str,
        script_file: str
):
    """
    更新迁移信息。
    """
    logger.debug(
        'code=%s, source_repository_code=%s, target_repository_code=%s, target_storage_code=%s, script_file=%s',
        code, source_repository_code,
        target_repository_code, target_storage_code,
        script_file
    )
    ensure(code is not None, '缺少迁移名称')
    ensure(
        source_repository_code is not None
        or (target_repository_code is not None and target_storage_code is not None)
        or script_file is not None,
        '缺少更新内容'
    )

    transformer = transformer_find(code)
    ensure(transformer is not None, f'迁移({code})不存在')

    if source_repository_code is not None:
        repository = repository_service.repository_find(source_repository_code)
        ensure(repository is not None, f'来源仓库({source_repository_code})不存在')
        source_id = repository.id
    else:
        source_id = transformer.source_id

    if target_repository_code is not None and target_storage_code is not None:
        storage = storage_service.storage_find(target_repository_code, target_storage_code)
        ensure(storage is not None, f'目标仓库({target_repository_code})的存储单元({target_storage_code})不存在')
        target_id = storage.id
    else:
        target_id = transformer.target_id

    if script_file is not None:
        script_file = absolute_path(script_file)
    else:
        script_file = transformer.script_file

    with metadata.connect() as meta:
        transformer = meta.record('''
          update
            transformers
          set
            source_id = ?,
            target_id = ?,
            script_file = ?,
            updated_at = datetime(current_timestamp, 'localtime')
          where
            code = ?
          returning *
        ''', source_id, target_id, script_file, code, constructor=Transformer._make)
        logger.info('更新迁移(%s)', code)
        logger.debug('transformer=%s', transformer)


def transformer_delete(code: str):
    """
    删除迁移。
    """
    logger.debug('code=%s', code)
    ensure(transformer_exists(code), f'迁移({code})不存在')
    with metadata.connect() as meta:
        transformer = meta.record('''
          delete from transformers where code = ? returning *
        ''', code, constructor=Transformer._make)
        logger.info('删除迁移(%s)', code)
        logger.debug('transformer=%s', transformer)


def transformer_list(
        source_repository_kind: Optional[str] = None,
        source_repository_code: Optional[str] = None,
        target_repository_kind: Optional[str] = None,
        target_repository_code: Optional[str] = None,
        target_storage_code: Optional[str] = None,
) -> list[TransformerProjection]:
    """
    列出所有迁移。
    """
    logger.debug(
        'source_repository_kind=%s, source_repository_code=%s, target_repository_kind=%s, target_repository_code=%s, target_storage_code=%s',
        source_repository_kind, source_repository_code,
        target_repository_kind, target_repository_code,
        target_storage_code,
    )

    filters = []
    parameters = []
    if source_repository_kind is not None:
        RepositoryKind.ensure(source_repository_kind)
        filters.append('and sources.kind = ?')
        parameters.append(source_repository_kind)
    if source_repository_code is not None:
        filters.append('and sources.code = ?')
        parameters.append(source_repository_code)
    if target_repository_kind is not None:
        RepositoryKind.ensure(target_repository_kind)
        filters.append('and targets.kind = ?')
        parameters.append(target_repository_kind)
    if target_repository_code is not None:
        filters.append('and targets.code = ?')
        parameters.append(target_repository_code)
    if target_storage_code is not None:
        filters.append('and storages.code = ?')
        parameters.append(target_storage_code)

    with metadata.connect() as meta:
        return meta.records(f'''
          with transformers_tasks as (
            select
              transformer_id as id,
              count(*) as tasks
            from
              tasks_transformers
            group by
              transformer_id
          )
          select
            transformers.code,
            sources.kind as source_repository_kind,
            sources.code as source_repository_code,
            targets.kind as target_repository_kind,
            targets.code as target_repository_code,
            storages.code as target_storage_code,
            transformers.script_file,
            coalesce(transformers_tasks.tasks, 0) as tasks
          from
            transformers
          inner join
            repositories as sources
          on
            transformers.source_id = sources.id
          inner join
            storages
          on
            transformers.target_id = storages.id
          inner join
            repositories as targets
          on
            storages.repository_id = targets.id
            {' '.join(filters)}
          left join
            transformers_tasks
          on
            transformers.id = transformers_tasks.id
          order by
            transformers.code
        ''', *parameters, constructor=TransformerProjection._make)


# 执行迁移

def transformer_execute(code: str):
    """
    执行迁移。
    """
    logger.debug('code=%s', code)
    transformer = transformer_find(code)
    ensure(transformer is not None, f'迁移({code})不存在')
    ensure(exists(transformer.script_file), f'迁移脚本({transformer.script_file})不存在')
    sql = slurp(transformer.script_file)

    with metadata.connect() as meta:
        context = meta.record('''
          select
            sources.code as source_repository_code,
            targets.code as target_repository_code,
            storages.code as target_storage_code
          from
            transformers
          inner join
            repositories as sources
          on
            transformers.source_id = sources.id
            and transformers.id = ?
          inner join
            storages
          on
            transformers.target_id = storages.id
          inner join
            repositories as targets
          on
            storages.repository_id = targets.id
        ''', transformer.id, constructor=TransformContext._make)
        source_repository = repository_service.repository_find(context.source_repository_code)
        target_repository = repository_service.repository_find(context.target_repository_code)
        target_storage = storage_service.storage_find(context.target_repository_code, context.target_storage_code)
        kind = RepositoryKind.of(target_repository.kind)

        with source_repository.connect() as source_connection:
            with source_connection.prepare(sql) as statement:
                kind.transform(statement, target_repository, target_storage)
                logger.info('从仓库(%s)迁移数据到仓库(%s)的存储单元(%s)', source_repository.code, target_repository.code, target_storage.code)



"""
数据迁移至多维表格，原理如下：
1. 在来源仓库上执行SQL。
2. 对比查询结果与快照是否一致。
3. 若快照已失效，则清空多维表格的记录。
4. 清空多维表格后，再添加新记录到多维表格。
5. 记录保存完成后，讲记录编码保存至本地缓存。
"""
import logging
from typing import Any

from duckcp.entity.statement import Statement
from duckcp.entity.storage import Storage
from duckcp.feishu.bitable import batch_delete, batch_create, Record
from duckcp.helper.digest import sha256
from duckcp.repository.bitable_repository import BiTableRepository
from duckcp.service import snapshot_service

logger = logging.getLogger(__name__)


def digest(records: list[dict[str, Any]]) -> str:
    """
    计算记录的摘要。
    """
    content = '\n'.join([
        f'{name}:{value}'
        for record in records
        for name, value in sorted(record.items())
    ])
    return sha256(content)


def bitable_transform(statement: Statement, repository: BiTableRepository, storage: Storage):
    """
    将数据源迁移到多维表格中。
    """
    authenticator = repository.authenticator
    logger.debug('storage=%s', storage)
    document = storage.properties['document']
    table = storage.properties['table']
    logger.debug('document=%s, table=%s', document, table)

    # 1. 获取数据并计算摘要
    records = statement.all()
    checksum = digest(records)
    logger.debug('records=%s, checksum=%s', records, checksum)

    # 2. 对比快照
    snapshot = snapshot_service.snapshot_find(storage.id)
    if snapshot is not None and snapshot.checksum != checksum and bool(snapshot.records):
        # 快照失效，先清空历史数据
        batch_delete(authenticator(), document, table, snapshot.records)
        logger.info('清空飞书文档(%s)多维表格(%s)', document, table)

    # 3. 保存数据
    if snapshot is None or snapshot.checksum != checksum and bool(records):
        # 无快照或快照失效，则添加数据
        records = batch_create(authenticator(), document, table, [
            Record(fields=record)
            for record in records
        ])
        snapshot_service.take_snapshot(storage.id, checksum, [record['record_id'] for record in records])
        logger.info('保存飞书文档(%s)多维表格(%s)记录%s条', document, table, len(records))
    else:
        logger.info('飞书文档(%s)多维表格(%s)数据未变化', document, table)



import logging
from contextlib import contextmanager
from os import makedirs
from os.path import dirname, exists
from typing import Optional, Iterator

from duckcp.configuration import Configuration
from duckcp.constant import IDENTIFIER
from duckcp.entity.executor import Executor
from duckcp.helper.fs import absolute_path, config
from duckcp.repository.sqlite_repository import SqliteRepository

logger = logging.getLogger(__name__)


def enable_metadata_configuration(file: Optional[str] = None):
    """
    元数据配置。
    """
    file = absolute_path(file) if file is not None else config(IDENTIFIER, 'configuration.db')
    folder = dirname(file)
    if not exists(folder):
        logger.info('创建目录(%s)', folder)
        makedirs(folder)
    Configuration.file = file
    logger.debug('元信息文件(%s)', file)


@contextmanager
def connect() -> Iterator[Executor]:
    """
    链接元信息数据库。
    """
    repository = SqliteRepository(properties={'file': Configuration.file})
    with repository.connect() as connection:
        with connection.executor() as executor:
            yield executor



"""
数据迁移至本地文件，原理如下：
1. 在来源仓库上执行SQL，并将查询结果封装成DataFrame。
2. 将DataFrame映射成DuckDB的只读视图。
3. 执行`COPY ... to ...`导出数据到本地文件。
"""
import logging

from duckcp.entity.statement import Statement
from duckcp.entity.storage import Storage
from duckcp.helper.fs import WorkDirectory
from duckcp.helper.sql import copy_to
from duckcp.repository.duckdb_repository import DuckDBRepository

logger = logging.getLogger(__name__)


def file_transform(statement: Statement, repository: DuckDBRepository, storage: Storage):
    """
    将数据源迁移到本地文件中。
    """
    folder = repository.properties['folder']
    with WorkDirectory(folder):
        with repository.establish_connection() as connection:
            with connection.cursor() as cursor:
                data = statement()
                cursor.execute(f' set global pandas_analyze_sample = {len(data)} ')
                cursor.register(storage.code, data)  # 表的全名是`temp.main.<table>`
                file_name = storage.properties.pop('file')
                # 通过sqlglot生成SQL，避免字符串转义或SQL注入等问题。
                ast = copy_to(storage.code, file_name, storage.properties)
                sql = ast.sql(dialect='duckdb')
                logger.debug('sql=%s', sql)
                cursor.execute(sql)



from typing import NamedTuple


class TaskTransformerProjection(NamedTuple):
    task_code: str  # 任务编码
    sort: int  # 执行顺序
    transformer_code: str  # 迁移编码
    source_repository_kind: str  # 来源仓库类型
    source_repository_code: str  # 来源仓库编码
    target_repository_kind: str  # 目标仓库类型
    target_repository_code: str  # 目标仓库编码
    target_storage_code: str  # 目标存储编码
    script_file: str  # 迁移脚本



import logging
from collections import namedtuple
from typing import Self, Any, Sequence, Optional

from pandas import DataFrame

from duckcp.typing.cursor_protocol import CursorProtocol
from duckcp.typing.record_constructor_protocol import RecordConstructorProtocol

logger = logging.getLogger(__name__)


class Executor:
    """
    执行器：执行SQL语句。
    """
    cursor: CursorProtocol

    def __init__(self, cursor: CursorProtocol):
        self.cursor = cursor

    def __enter__(self) -> Self:
        """
        生命周期开始。
        """
        return self

    def __exit__(self, exception_class, exception, traceback):
        """
        生命周期结束。
        """
        self.close()

    def close(self):
        """
        关闭本次会话/游标。
        """
        self.cursor.close()

    def batch(self, sql: str, parameters: list[Sequence[Any]]):
        """
        批量执行语句，无返回结果。
        """
        logger.debug('sql=%s, parameters=%s', sql, parameters)
        self.cursor.executemany(sql, parameters)

    def execute(self, sql: str, *parameters: Any) -> tuple[list[str], list[Sequence[Any]]]:
        """
        执行单条语句，返回原始的结果。
        - 头信息：列名和类型。
        - 记录：原始数据。
        """
        logger.debug('sql=%s, parameters=%s', sql, parameters)
        if parameters:
            self.cursor.execute(sql, parameters)
        else:
            self.cursor.execute(sql)  # 勿删：不同类型Cursor中参数默认值不同，无法统一处理
        columns = [column[0] for column in self.cursor.description] if self.cursor.description else []
        records = self.cursor.fetchall()
        return columns, records

    def __call__(self, sql: str, *parameters: Any) -> DataFrame:
        """
        执行查询语句，返回DataFrame结构。
        """
        columns, records = self.execute(sql, *parameters)
        return DataFrame(records, columns=columns)

    def records[T: tuple[Any, ...]](self, sql: str, *parameters: Any, constructor: RecordConstructorProtocol[T] = None) -> list[T]:
        """
        执行查询语句，返回NamedTuple形式的记录数据。
        """
        columns, records = self.execute(sql, *parameters)
        if constructor is None:
            record_class = namedtuple('_DuckCP_Record', columns)
            return [record_class(*record) for record in records]
        else:
            return [constructor(record) for record in records]

    def record[T: tuple[Any, ...]](self, sql: str, *parameters: Any, constructor: RecordConstructorProtocol[T] = None) -> Optional[T]:
        """
        执行查询语句，并返回第一行NamedTuple形式的记录数据。
        """
        records = self.records(sql, *parameters, constructor=constructor)
        return records[0] if records else None

    def value(self, sql: str, *parameters: Any) -> Optional[Any]:
        """
        执行查询语句，并返回第一行第一列数据。
        """
        record = self.record(sql, *parameters)
        return record[0] if record else None

    def values(self, sql: str, *parameters: Any) -> list[Any]:
        """
        执行查询语句，并返回所有行的第一列数据。
        """
        return [record[0] for record in self.records(sql, *parameters)]

    def all(self, sql: str, *parameters: Any) -> list[dict[str, Any]]:
        """
        执行查询语句，返回字典结构的记录。
        """
        columns, records = self.execute(sql, *parameters)
        return [
            {column: value for column, value in zip(columns, record)}
            for record in records
        ]

    def one(self, sql: str, *parameters: Any) -> dict[str, Any]:
        """
        执行查询语句，返回字典结构的第一行记录。
        """
        records = self.all(sql, *parameters)
        return records[0] if records else None



from datetime import datetime
from typing import NamedTuple


class Credential(NamedTuple):
    id: int
    platform_code: str  # 所属平台
    app_code: str  # 应用编码
    access_token: str  # 访问凭证
    expired_at: datetime
    created_at: datetime
    updated_at: datetime



import logging
from collections.abc import Iterator
from contextlib import contextmanager

import duckdb
from duckdb.duckdb import DuckDBPyConnection

from duckcp.entity.connection import Connection
from duckcp.entity.repository import Repository
from duckcp.helper.fs import WorkDirectory
from duckcp.helper.validation import ensure

logger = logging.getLogger(__name__)


class FileRepository(Repository):
    """
    文件类型仓库。
    """

    def establish_connection(self) -> DuckDBPyConnection:
        """
        创建DuckDB内存数据库连接。
        """
        return duckdb.connect(':memory:')

    @contextmanager
    def connect(self) -> Iterator[Connection]:
        """
        建立在目标目录下的临时连接。
        """
        ensure(bool(self.properties), '缺少连接参数')
        ensure(bool(self.properties.get('folder')), '缺少文件夹')
        folder = self.properties.get('folder')
        with WorkDirectory(folder):
            yield Connection(self.establish_connection())



from abc import abstractmethod
from datetime import datetime
from typing import NamedTuple, Any

from duckcp.entity.connection import Connection
from duckcp.typing.connection_protocol import ConnectionProtocol


class Repository(NamedTuple):
    id: int = None
    kind: str = None  # 类型
    code: str = None  # 编码
    properties: dict[str, Any] = None  # 连接信息
    created_at: datetime = None
    updated_at: datetime = None

    @abstractmethod
    def establish_connection[T: ConnectionProtocol](self) -> T:
        """
        建立新的数据库连接，并返回原始的连接类型。
        """
        pass

    def connect(self) -> Connection:
        """
        建立新的数据库连接。
        """
        return Connection(self.establish_connection())



from typing import Self

from duckcp.entity.executor import Executor
from duckcp.entity.statement import Statement
from duckcp.typing.connection_protocol import ConnectionProtocol


class Connection:
    """
    数据库连接对象。
    """
    connection: ConnectionProtocol

    def __init__(self, connection: ConnectionProtocol):
        self.connection = connection

    def __enter__(self) -> Self:
        """
        生命周期开始。
        """
        return self

    def __exit__(self, exception_class, exception, traceback):
        """
        生命周期结束。
        """
        self.close()

    def close(self):
        """
        断开连接。
        """
        self.connection.close()

    def executor(self) -> Executor:
        """
        创建新的语句对象，对于执行查询语句。
        """
        return Executor(self.connection.cursor())

    def prepare(self, sql: str) -> Statement:
        """
        准备SQL语句用于后续执行。
        """
        return Statement(self.executor(), sql)



import logging
from collections.abc import Sequence, Mapping
from numbers import Number
from typing import Optional, Any

from sqlglot import parse, Expression
from sqlglot.dialects.duckdb import DuckDB
from sqlglot.expressions import With, CTE, Table, Create, Identifier, From, Delete, Insert, Schema, Values, Tuple, Copy, Literal, CopyParameter, Var, Boolean, Struct, Array, Null, PropertyEQ

logger = logging.getLogger(__name__)


def to_expression(instance: Any) -> Expression:
    """
    将Python值转成表达式。
    """
    if instance is None:
        return Null()
    elif isinstance(instance, bool):
        return Boolean(this=instance)
    elif isinstance(instance, str):
        return Literal.string(instance)
    elif isinstance(instance, Number):
        return Literal.number(instance)
    elif isinstance(instance, Sequence):
        return Array(expressions=[to_expression(value) for value in instance])
    elif isinstance(instance, Mapping):
        return Struct(expressions=[PropertyEQ(
            this=Identifier(this=name, quoted=True),
            expression=to_expression(value),
        ) for name, value in instance.items()])
    else:
        raise ValueError(f'未知数据({instance})类型({type(instance)})')


def extract_tables(sql: str) -> set[str]:
    """
    从SQL中提取真正的表名，忽略CTE、子查询等临时的表名。
    """
    logger.debug('sql=%s', sql)
    tables = set()  # 真正的表
    for statement in parse(sql, dialect=DuckDB) or []:
        ignores = set()  # 提取CTE别名，忽略这些名字
        if expression := statement.find(With):
            for cte in expression.find_all(CTE):  # 先依次遍历CTE
                for table in cte.find_all(Table):
                    if table.name not in ignores:
                        tables.add(table.name)
                ignores.add(cte.alias)
        for expression in statement.iter_expressions():
            if not isinstance(expression, With):  # 排除With语句；前面已处理
                for table in expression.find_all(Table):
                    if table.name not in ignores:
                        tables.add(table.name)
    return tables


def create_or_replace_table(
        catalog: Optional[str],
        schema: Optional[str],
        table: str,
        source: str,
) -> Expression:
    """
    创建DuckDB方言的create or replace table语句。
    """
    logger.debug('catalog=%s, schema=%s, table=%s, source=%s', catalog, schema, table, source)
    return Create(
        this=Table(
            this=Identifier(this=table, quoted=True),
            db=Identifier(this=schema, quoted=True) if schema else None,
            catalog=Identifier(this=catalog, quoted=True) if catalog else None),
        kind='table',
        replace=True,
        expression=From(
            this=Table(
                this=Identifier(this=source, quoted=True),
                db=Identifier(this='main', quoted=False),
                catalog=Identifier(this='temp', quoted=False))))


def delete_from(
        catalog: Optional[str],
        schema: Optional[str],
        table: str,
) -> Expression:
    """
    创建通用的删除表所有数据语句。
    """
    logger.debug('catalog=%s, schema=%s, table=%s', catalog, schema, table)
    return Delete(
        this=Table(
            this=Identifier(this=table, quoted=True),
            db=Identifier(this=schema, quoted=True) if schema else None,
            catalog=Identifier(this=catalog, quoted=True) if catalog else None))


def insert_into(
        catalog: Optional[str],
        schema: Optional[str],
        table: str,
        columns: list[str],
) -> Expression:
    """
    创建通用的新增记录语句。
    """
    logger.debug('catalog=%s, schema=%s, table=%s, columns=%s', catalog, schema, table, columns)
    return Insert(
        this=Schema(
            this=Table(
                this=Identifier(this=table, quoted=True),
                db=Identifier(this=schema, quoted=True) if schema else None,
                catalog=Identifier(this=catalog, quoted=True) if catalog else None),
            expressions=[Identifier(this=column, quoted=True) for column in columns]),
        expression=Values(
            expressions=[Tuple(
                expressions=['?'] * len(columns)
            )]))


def copy_to(
        table_name: str,
        file_name: str,
        parameters: dict[str, Any],
) -> Expression:
    """
    创建DuckDB方言的COPY语句。
    """
    return Copy(
        this=Table(
            this=Identifier(this=table_name, quoted=True),
            db=Identifier(this='main', quoted=True),
            catalog=Identifier(this='temp', quoted=True)),
        files=[Literal.string(file_name)],
        params=[
            CopyParameter(this=Var(this=name), expression=to_expression(value))
            for name, value in parameters.items()
        ])



"""
开放平台认证服务。
"""
import logging
from typing import Optional

from duckcp.configuration import meta_configuration as metadata
from duckcp.entity.credential import Credential
from duckcp.helper.validation import ensure
from duckcp.typing.authentication_token_type import AuthenticationToken
from duckcp.typing.authenticator_type import Authenticator
from duckcp.typing.credential_refresher_type import CredentialRefresher

logger = logging.getLogger(__name__)


def authenticate(platform_code: str, app_code: str, token: AuthenticationToken, refresher: CredentialRefresher) -> Authenticator:
    """
    认证并获取最新的访问凭证。
    """
    logger.debug('platform_code=%s, app_code=%s, token=%s', platform_code, app_code, token)
    ensure(platform_code is not None and app_code is not None and token is not None, '缺少应用信息')

    def authenticator() -> Optional[str]:
        """
        认证闭包：包含了认证信息。
        """
        with metadata.connect() as meta:
            access_token = meta.value('''
              select
                access_token
              from
                credentials
              where
                platform_code = ?
                and app_code = ?
                and datetime(current_timestamp, 'localtime') < expired_at
            ''', platform_code, app_code)
            if access_token is None:
                access_token, expired_at = refresher(token)
                credential = meta.record('''
                  insert or replace into credentials
                    (platform_code, app_code, access_token, expired_at)
                  values
                    (?, ?, ?, ?)
                  returning *
                ''', platform_code, app_code, access_token, expired_at, constructor=Credential._make)
                logger.info('刷新凭证(%s, %s)', platform_code, app_code)
                logger.debug('credential=%s', credential)
            return access_token

    return authenticator



import logging

from click import help_option, argument, option, INT
from rich.console import Console
from rich.table import Table

from duckcp.boot import app
from duckcp.service import task_service

logger = logging.getLogger(__name__)


@app.group(short_help='管理任务', help='管理任务定时调度的任务')
@help_option('-h', '--help', help='展示帮助信息')
def task():
    pass


@task.command('create', help='创建任务')
@argument('name', metavar='NAME')
@help_option('-h', '--help', help='展示帮助信息')
def task_create(name: str):
    logger.debug('name=%s', name)
    task_service.task_create(name)


@task.command('delete', help='删除任务')
@argument('name', metavar='NAME')
@help_option('-h', '--help', help='展示帮助信息')
def task_delete(name: str):
    logger.debug('name=%s', name)
    task_service.task_delete(name)


@task.command('list', help='列出所有任务')
@help_option('-h', '--help', help='展示帮助信息')
def task_list():
    table = Table(title='任务列表')
    table.add_column('名称', no_wrap=True)
    table.add_column('关联迁移', justify='right')
    for row in task_service.task_list():
        table.add_row(
            row.code,
            str(row.transformers),
        )

    console = Console()
    console.print(table)


@task.command('execute', help='执行任务')
@argument('name', metavar='NAME')
@help_option('-h', '--help', help='展示帮助信息')
def task_execute(name: str):
    logger.debug('name=%s', name)
    task_service.task_execute(name)


@task.command('bind', help='绑定迁移')
@argument('name', metavar='NAME')
@option('-t', '--transformer', metavar='TRANSFORMER', required=True, help='关联迁移')
@option('-i', '--index', metavar='NUMBER', type=INT, help='执行顺序；默认最后')
@help_option('-h', '--help', help='展示帮助信息')
def task_bind(name: str, transformer: str, index: int):
    logger.debug('name=%s, transformer=%s', name, transformer)
    task_service.task_bind(name, transformer, index)


@task.command('unbind', help='解绑迁移')
@argument('name', metavar='NAME')
@option('-t', '--transformer', metavar='TRANSFORMER', required=True, help='关联迁移')
@help_option('-h', '--help', help='展示帮助信息')
def task_unbind(name: str, transformer: str):
    logger.debug('name=%s, transformer=%s', name, transformer)
    task_service.task_unbind(name, transformer)


@task.command('transformers', help='列出所有迁移')
@argument('name', required=False, metavar='NAME')
@help_option('-h', '--help', help='展示帮助信息')
def task_transformer_list(name: str):
    logger.debug('name=%s', name)
    table = Table(title='任务迁移列表')
    if name is None:
        table.add_column('任务名称', no_wrap=True)
    table.add_column('执行顺序', justify='right')
    table.add_column('来源仓库类型')
    table.add_column('来源仓库编码')
    table.add_column('目标仓库类型')
    table.add_column('目标仓库编码')
    table.add_column('目标存储编码')
    table.add_column('迁移脚本')
    for row in task_service.task_transformer_list(name):
        record = [
            str(row.sort),
            row.source_repository_kind,
            row.source_repository_code,
            row.target_repository_kind,
            row.target_repository_code,
            row.target_storage_code,
            row.script_file
        ]
        if name is None:
            record.insert(0, row.task_code)
        table.add_row(*record)

    console = Console()
    console.print(table)



from datetime import datetime
from typing import NamedTuple


class Transformer(NamedTuple):
    id: int
    code: str  # 编号
    source_id: int  # 来源仓库
    target_id: int  # 目标表格
    script_file: str  # 迁移脚本
    created_at: datetime
    updated_at: datetime



from datetime import datetime
from typing import NamedTuple, Any


class Snapshot(NamedTuple):
    id: int
    storage_id: int  # 所属存储单元
    checksum: str  # 摘要
    records: list[Any]  # 记录
    created_at: datetime
    updated_at: datetime



import click
from click import Choice, Tuple

from duckcp.configuration.logging_configuration import enable_logging_configuration
from duckcp.configuration.meta_configuration import enable_metadata_configuration


@click.group(help='数据同步工具')
@click.option('-c', '--config-file', metavar='FILE', help='配置文件')
@click.option('-o', '--logging-file', metavar='FILE', help='日志文件')
@click.option('-l', '--logging', type=Tuple([str, Choice([
    'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL',
], case_sensitive=False)]), metavar='<NAME LEVEL>...', multiple=True, nargs=2, help='日志等级')
@click.option('-m', '--message-only', is_flag=True, help='日志只输出内容')
@click.option('-v', '--verbose', is_flag=True, help='开启详细日志')
@click.option('-q', '--quiet', is_flag=True, help='关闭所有日志')
@click.version_option('v0.1.2', '-V', '--version', help='展示版本信息')
@click.help_option('-h', '--help', help='展示帮助信息')
def app(config_file: str, logging_file: str, logging: list[tuple[str, str]], message_only: bool, verbose: bool, quiet: bool) -> None:
    enable_logging_configuration(logging_file, logging, message_only, not quiet and verbose, quiet)
    enable_metadata_configuration(config_file)



from typing import Protocol, Sequence, Any

from duckcp.typing.supports_get_item_protocol import SupportsGetItemProtocol


class CursorProtocol(Protocol):
    """
    一个抽象基类，定义数据库游标协议。
    """

    @property
    def description(self) -> Sequence[SupportsGetItemProtocol]:
        ...

    def close(self):
        """
        关闭游标
        """
        ...

    def executemany(self, sql: str, parameters: list[Sequence[Any]]):
        """
        批量执行。
        """
        ...

    def execute(self, sql: str, parameters: Sequence[Any] = None):
        """
        单句执行。
        """
        ...

    def fetchall(self) -> list[Sequence[Any]]:
        """
        获取查询结果。
        """
        ...



import logging
from typing import Any, Sequence, cast, NamedTuple

import duckdb
from duckdb.duckdb import DuckDBPyConnection
from pandas import DataFrame

from duckcp.configuration import meta_configuration as metadata
from duckcp.entity.connection import Connection
from duckcp.entity.executor import Executor
from duckcp.entity.repository import Repository
from duckcp.feishu import tenant_access_token, bitable
from duckcp.helper.sql import extract_tables
from duckcp.helper.validation import ensure
from duckcp.service.authentication_service import Authenticator, authenticate
from duckcp.typing.connection_protocol import ConnectionProtocol
from duckcp.typing.supports_get_item_protocol import SupportsGetItemProtocol

logger = logging.getLogger(__name__)


class BiTable(NamedTuple):
    name: str  # 在SQL中使用的表名
    code: str  # 飞书多维表格编码
    document_code: str  # 飞书多维文档编码


class BiTableCursor:
    """
    飞书多维表格游标：增删改等变更操作只影响本地缓存数据；不会同步至多维表格。
    """
    cursor: duckdb.DuckDBPyConnection
    authenticator: Authenticator  # 授信服务。
    tables: dict[str, BiTable]

    def __init__(self, cursor: DuckDBPyConnection, authenticator: Authenticator, tables: dict[str, BiTable]):
        self.cursor = cursor
        self.authenticator = authenticator
        self.tables = tables

    @property
    def description(self) -> Sequence[SupportsGetItemProtocol]:
        return self.cursor.description

    def close(self):
        """
        关闭游标
        """
        self.cursor.close()

    def __prepare(self, sql: str):
        """
        根据SQL语句中查询用到的表，动态地从远程多维表格中实时加载最新的数据。
        """
        for table_name in extract_tables(sql):
            if table := self.tables.get(table_name):
                records = bitable.list_records(self.authenticator(), table.document_code, table.code)
                frame = DataFrame([{
                    **record['fields'],
                    'id': record['record_id'],
                } for record in records])
                self.cursor.execute(f' set global pandas_analyze_sample = {len(frame)} ')
                self.cursor.register(table.name, frame)

    def executemany(self, sql: str, parameters: list[Sequence[Any]]):
        """
        批量执行。
        """
        self.__prepare(sql)
        self.cursor.executemany(sql, parameters)

    def execute(self, sql: str, parameters: Sequence[Any]):
        """
        单句执行。
        """
        self.__prepare(sql)
        return self.cursor.execute(sql, parameters)

    def fetchall(self) -> list[Sequence[Any]]:
        """
        获取查询结果。
        """
        return self.cursor.fetchall()


class BiTableConnection(Connection):
    """
    飞书多维表格连接：获取访问凭证。
    """
    authenticator: Authenticator  # 授信服务。
    tables: dict[str, BiTable]

    def __init__(self, connection: ConnectionProtocol, authenticator: Authenticator, tables: dict[str, BiTable]):
        super().__init__(connection)
        self.authenticator = authenticator
        self.tables = tables

    def executor(self) -> Executor:
        """
        创建新的语句对象，对于执行查询语句。
        """
        cursor = cast(DuckDBPyConnection, self.connection.cursor())
        return Executor(BiTableCursor(cursor, self.authenticator, self.tables))


class BiTableRepository(Repository):
    """
    飞书多维表格类型仓库。
    """

    def establish_connection(self) -> DuckDBPyConnection:
        """
        创建DuckDB内存数据库连接。
        """
        return duckdb.connect(':memory:')

    @property
    def authenticator(self) -> Authenticator:
        """
        飞书开放平台的身份校验器。
        """
        ensure(bool(self.properties), '缺少连接参数')
        ensure(bool(self.properties.get('access_key')), '缺少访问编码')
        access_key = self.properties.get('access_key')
        ensure(bool(self.properties.get('access_secret')), '缺少访问密钥')
        access_secret = self.properties.get('access_secret')
        logger.debug('access_key=%s', access_key)
        return authenticate('feishu', access_key, {
            'access_key': access_key,
            'access_secret': access_secret,
        }, tenant_access_token)

    def connect(self) -> Connection:
        """
        连接多维表格。
        """
        tables = {}
        if self.id is not None:
            with metadata.connect() as meta:
                tables = {
                    table.name: table
                    for table in meta.records('''
                      select
                        code as name,
                        properties->>'table' as code,
                        properties->>'document' as document_code
                      from
                        storages
                      where
                        repository_id = ?
                    ''', self.id, constructor=BiTable._make)
                }
        connection = self.establish_connection()
        return BiTableConnection(connection, self.authenticator, tables)



import logging
from typing import Optional

from duckcp.configuration import meta_configuration as metadata
from duckcp.entity.snapshot import Snapshot

logger = logging.getLogger(__name__)


def snapshot_find(storage_id: int) -> Optional[Snapshot]:
    """
    找到存储单元的快照记录。
    """
    with metadata.connect() as meta:
        return meta.record('select * from snapshots where storage_id = ?', storage_id, constructor=Snapshot._make)


def take_snapshot(storage_id: int, checksum: str, records: list[str]):
    """
    保存快照。
    """
    with metadata.connect() as meta:
        snapshot = meta.record('''
          insert or replace into snapshots
            (storage_id, checksum, records)
          values
            (?, ?, ?)
          returning *
        ''', storage_id, checksum, records, constructor=Snapshot._make)
        logger.info('创建快照(%s)', storage_id)
        logger.debug('snapshot=%s', snapshot)
